import os
import re
import glob
import numpy as np
import pandas as pd
from datatable import Frame
from PharmacoDI.combine_pset_tables import write_table


# TODO: add docstrings

def get_metadata(file_name):
    # Check that metadata file exists
    if not os.path.exists(file_name):
        raise ValueError(
            f'No metadata file {file_name} could be found!')
    
    # Read csv file and return df
    return pd.read_csv(file_name)


# --- SYNONYMS TABLES --------------------------------------------------------------------------

def build_cell_synonym_df(cell_file, output_dir):
    # Get metadata file and cell_df
    cell_metadata = get_metadata(cell_file)
    cell_df = pd.read_csv(os.path.join(output_dir, 'cell.csv'))
    # Find all columns relevant to cellid
    pattern = re.compile('cellid')
    cell_columns = cell_metadata[[
        col for col in cell_metadata.columns if pattern.search(col)]]

    # Get all unique synonyms and join with cell_df
    cell_synonym_df = melt_and_join(cell_columns, 'unique.cellid', cell_df)
    cell_synonym_df = cell_synonym_df.rename(columns={'id': 'cell_id', 'value': 'cell_name'})

    # Add blank col for dataset_id (TODO)
    cell_synonym_df['dataset_id'] = np.nan

    # Convert to datatable.Frame for fast write to disk
    df = Frame(cell_synonym_df)
    df = write_table(df, 'cell_synonym', output_dir)
    return df
    

def build_tissue_synonym_df(tissue_file, output_dir):
    # Get metadata file and tissue_df (assume taht tissue_df is also in output_dir)
    tissue_metadata = get_metadata(tissue_file)
    tissue_df = pd.read_csv(os.path.join(output_dir, 'tissue.csv'))

    # Find all columns relevant to tissueid
    pattern = re.compile('tissueid')
    tissue_cols = tissue_metadata[[
        col for col in tissue_metadata.columns if pattern.search(col)]]

    # Get all unique synonyms and join with tissue_df
    tissue_synonym_df = melt_and_join(tissue_cols, 'unique.tissueid', tissue_df)
    tissue_synonym_df = tissue_synonym_df.rename(columns={'id': 'tissue_id', 'value': 'tissue_name'})

    # Add blank col for dataset_id (TODO)
    tissue_synonym_df['dataset_id'] = np.nan

    # Convert to datatable.Frame for fast write to disk
    df = Frame(tissue_synonym_df)
    df = write_table(df, 'tissue_synonym', output_dir)
    return df


def build_compound_synonym_df(compound_file, output_dir):
    # Get metadata file and compound_df
    compound_metadata = get_metadata(compound_file)
    compound_df = pd.read_csv(os.path.join(output_dir, 'compound.csv'))

    # Find all columns relevant to compoundid
    # Right now only FDA col is dropped, but may be more metadata in the future
    pattern = re.compile('drugid')
    compound_cols= compound_metadata[[
        col for col in compound_metadata.columns if pattern.search(col)]]

    # Get all unique synonyms and join with compounds_df
    compound_synonym_df = melt_and_join(compound_cols, 'unique.drugid', compound_df)
    compound_synonym_df = compound_synonym_df.rename(columns={'id': 'compound_id', 'value': 'compound_name'})

    # Add blank col for dataset_id (TODO)
    compound_synonym_df['dataset_id'] = np.nan

    # Convert to datatable.Frame for fast write to disk
    df = Frame(compound_synonym_df)
    df = write_table(df, 'compound_synonym', output_dir)
    return df


# Helper function for getting all synonyms related to a certain df
def melt_and_join(meta_df, unique_id, join_df):
    """
    @param meta_df: [`Dask DataFrame`] The DataFrame containing all the synonyms (metadata)
    @param unique_id: [`string`] The name of the column in the metadata containing the unique IDs
    @param join_df: [`Dask DataFrame`] THe DataFrame containing the primary keys that will be used as
        foreign keys in the new synonyms df

    @return [`DataFrame`] The synonys dataframe, with a PK, FK based on join_df, and all unique synonyms
    """
    # Convert wide meta_df to long table
    # Drop 'variable' col (leave only unique ID and synonyms), drop duplicates
    synonyms = pd.melt(meta_df, id_vars=[unique_id])[
        [unique_id, 'value']].drop_duplicates()

    # Drop all rows where value is NA
    synonyms = synonyms[synonyms['value'].notnull()]

    # Join with join_df based on unique_id
    synonyms = pd.merge(synonyms, join_df, left_on=unique_id,
                        right_on='name', how='inner')[['id', 'value']]
    synonyms['id'] = synonyms['id'].astype('int')

    return synonyms




import os
import glob
import re
import pandas as pd
import numpy as np
from datetime import date
from .build_primary_pset_tables import build_primary_pset_tables, build_cell_df, build_compound_df, build_tissue_df
from .build_experiment_tables import build_experiment_tables, build_experiment_df
from .build_gene_compound_tissue_dataset_tables import build_gene_compound_tissue_dataset_df
from .write_pset_table import write_pset_table
from .build_dataset_join_tables import build_dataset_join_dfs


def build_all_pset_tables(pset_dict, pset_name, procdata_dir, gene_sig_dir):
    """
    Build all tables for a dataset and write them to a directory of all processed data.

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @param pset_name: [`string`] The name of the PSet
    @param procdata_dir: [`string`] The file path to the directory containing processed data
    @param gene_sig_dir: [`string`] The file path to the directory containing gene_compounds data
    @return: [`None`]
    """
    pset_dfs = {}

    # Build primary tables (relating to cells, compounds, tissues, genes)
    print('Building primary tables...')
    pset_dfs = build_primary_pset_tables(pset_dict, pset_name)

    print('Building dataset join tables...')
    pset_dfs = {**pset_dfs, **build_dataset_join_dfs(
        pset_dict, pset_name, pset_dfs)}

    # Build experiment tables
    print('Building experiment tables...')
    # FIX: Modified to use pre-3.9 syntax to ensure backwards compatibility
    pset_dfs = {**pset_dfs, **build_experiment_tables(
        pset_dict, pset_name, pset_dfs['cell'])}

    # Build gene compounds table
    print('Building gene_compound_tissue_dataset table...')
    pset_dfs['gene_compound_tissue_dataset'] = build_gene_compound_tissue_dataset_df(gene_sig_dir, pset_name)
    if not isinstance(pset_dfs['gene_compound_tissue_dataset'], pd.DataFrame):
        del pset_dfs['gene_compound_tissue_dataset']

    # Build summary/stats tables
    print('Building mol_cell and dataset_stats tables...')
    if 'gene_compound_tissue_dataset' in pset_dfs:
        pset_dfs['mol_cell'] = build_mol_cell_df(
            pset_dict, pset_name, pset_dfs['gene_compound_tissue_dataset'], pset_dfs['dataset_cell'])
    pset_dfs['dataset_statistics'] = build_dataset_stats_df(
        pset_dict, pset_name, pset_dfs)

    # Write all tables to CSV files
    for df_name in pset_dfs.keys():
        write_pset_table(pset_dfs[df_name], df_name, pset_name, procdata_dir)

    log_file = open(os.path.join(procdata_dir, pset_name, f'{pset_name}_log.txt'), "w")
    log_file.write('Wrote the following PSet tables: \n')
    log_file.writelines(f'{table}\n' for table in pset_dfs.keys())
    log_file.write(f'on {date.today()}')
    log_file.close()


def build_mol_cell_df(pset_dict, pset_name, gene_compound_df, dataset_cell_df=None):
    """
    Builds a table that summarizes the number of profiles, per cell line, per molecular data
    type, in this dataset. (Only considers molecular data types for which there are sens stats?)

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @param gene_compound_df: [`pd.DataFrame`] The gene_compound table for this PSet
    @param dataset_cell_df: [`pd.DataFrame`] A table containing all the cells in this  
        PSet and the PSet name
    @return: [`pd.DataFrame`] The table with the number of profiles for each cell line, 
                                for each molecular data type
    """
    mol_cell_df = pd.DataFrame(
        columns=['cell_id', 'dataset_id', 'mDataType', 'num_prof'])

    molecularTypes = pd.unique(gene_compound_df['mDataType'])

    if 'molecularProfiles' in pset_dict:
        profiles_dict = pset_dict['molecularProfiles']
    else:
        profiles_dict = None

    if dataset_cell_df is None:
        dataset_cell_df = build_dataset_cell_df(
            pset_dict, pset_name, cell_df=None)

    for mDataType in molecularTypes:
        if isinstance(profiles_dict, dict):
            # Get the number of times each cellid appears in colData for that mDataType
            num_profiles = profiles_dict[mDataType]['colData']['cellid'].value_counts(
            )

            # Join with datasets cells on cellid
            df = pd.merge(dataset_cell_df, num_profiles,
                          left_on='cell_id', right_on=num_profiles.index, how='left')

            # Rename num_profiles column
            df.rename(columns={'cellid': 'num_prof'}, inplace=True)

            # Set mDataType column to the current molecular type
            df['mDataType'] = mDataType
        else:
            # If PSet contains no molecular profiles, set num_prof to 0
            # for all celll lines and all molecular data types
            df = dataset_cell_df.copy()
            df['mDataType'] = mDataType
            df['num_prof'] = 0

        # Append to mol_cell_df
        mol_cell_df = mol_cell_df.append(df)

    # Replace any NaN in the num_profiles column with 0
    mask = mol_cell_df.query('num_prof.isna()').index
    mol_cell_df.loc[mask, 'num_prof'] = 0
    mol_cell_df['num_prof'] = mol_cell_df['num_prof'].astype('int32')

    return mol_cell_df


def build_dataset_stats_df(pset_dict, pset_name, pset_dfs=None):
    """
    Summarizes how many cell lines, tissues, compounds, and experiments are contained
    within the dataset.

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @param pset_name: [`string`] The name of the PSet
    @param pset_dfs: [`dict`] A dictionary of tables from the PSet, with table names 
                                as the keys
    @return: [`pd.DataFrame`] A one-row table with the summary stats for this PSet
    """
    if pset_dfs is None:
        pset_dfs = {}
    if 'tissue' not in pset_dfs:
        pset_dfs['tissue'] = build_tissue_df(pset_dict)
    if 'cell' not in pset_dfs:
        pset_dfs['cell'] = build_cell_df(pset_dict)
    if 'compound' not in pset_dfs:
        pset_dfs['compound'] = build_compound_df(pset_dict)
    if 'experiment' not in pset_dfs:
        pset_dfs['experiment'] = build_experiment_df(
            pset_dict, pset_name, pset_dfs['cell'])

    return pd.DataFrame({
        'dataset_id': [pset_name],
        'cell_lines': [len(pset_dfs['cell'].index)],
        'tissues': [len(pset_dfs['tissue'].index)],
        'compounds': [len(pset_dfs['compound'].index)],
        'experiments': [len(pset_dfs['experiment'].index)]
    })



import sys
import os

def download_gene_signatures(user=None, password=None, remote="niagara.computecanada.ca", opts={},
                             remote_path="/scratch/b/bhaibeka/psmirnov/pearson_perm_res/", save_dir="data/rawdata/gene_signatures"):
    """
    Download all precomputed gene signatures from the `remote_path` directory on the `remote` server, excluding those
    already in `save_dir`.

    WARNING: The current implementation relies on rysnc being installed. Thus it will not work on Windows platforms.

    :param user [string] Your username for the remote server. In interactive sessions, if you exclude this argument
        you will be prompted to enter your username.
    :param password [string] Your password for the remote server. To avoid hard coding your password, we recommend
        you read this from an environmental variables using `os.environment`. In interactive sessions, if you exclude
        this argument you will be prompted to enter you password.
    :param remote: [string] Name or IP of remote server
    :param remote_path: [string] Path to the gene signature directory on the remote server
    :param save_dir: [string] Path to save the downloaded gene signature to.
    :return: [None] Syncs `save_dir` with `remote:remote_path`, downloading the most up to date gene signatures for
        PharmacoDB.
    """

    if user == None or password == None:
        if sys.flags.interactive:
            if user == None:
                user = input("Please enter your username for {}: ".format(remote))
            if password == None:
                password = input("Please enter your password for {}: ".format(remote))

        else:
            raise ValueError("You must pass `user` and `password` parameters to this function in non-interative Python sessions")

    os.system(f'rsync -hrtvP {user}@{remote}:{remote_path} {save_dir}')




import pandas as pd
import wget

def download_canonical_psets(save_dir, api_url= "https://www.orcestra.ca/api/psets/canonical"):
    """
    Download the specified canonical PSets from the specified api_url

    :param save_dir: [string] Path to save the PSets in
    :param api_url: [string] URL where available PSets can be retrieved. Defaults to current Orcestra API.
    :return: [None] Downloads PSet .rds files into save_dir using wget.
    """
    pset_df = pd.read_json(api_url)
    url_dict = pset_df.set_index('name').to_dict()['downloadLink']

    for name, url in url_dict.items():
        print("Downloading", name, "from", url, sep=" ")
        wget.download(url, save_dir + "/" + name + '.rds')

    return None



from chembl_webresource_client.new_client import new_client
import pandas as pd
import os


def get_chembl_targets(target_file):
    """
    Get all ChEMBL targets in humans and write them to a table.

    :target_file: full file path to where the target table should be written
    :return: a DataFrame containing all targets from ChEMBL
    """
    print('Getting all targets from ChEMBL...')
    # Initiate connection to target table
    target = new_client.target

    # Parse all human targets into a DataFrame
    target_result = target.filter(organism__in=['Homo sapiens'])
    results = list(target_result)
    target_df = pd.DataFrame(results)

    # Explode list-like columns into separate rows, duplicating the index
    object_columns = target_df.dtypes[target_df.dtypes ==
                                      'object'].index.values
    for column in object_columns:
        target_df = target_df.explode(column)

    # Drop any targets without cross refs
    target_df = target_df.query("cross_references.notna()").copy()

    # Expand cols with dtype dict into their cols for each key
    for col in ['cross_references', 'target_components']:
        dict_col = pd.json_normalize(target_df[col], max_level=0)
        dict_col.index = target_df.index
        target_df.drop(columns=col, inplace=True)
        target_df = pd.merge(target_df, dict_col,
                             left_index=True, right_index=True)

    # Drop target component cols with dicts (for now; TODO: keep them and also expand them?)
    target_df.drop(columns=['target_component_synonyms',
                            'target_component_xrefs'], inplace=True)
    target_df.drop_duplicates(inplace=True)

    target_df.to_csv(target_file)
    return target_df



import PharmacoDI as pdi

pset = 'GDSC_v1'
pset_dir = '../PharmacoDI_snakemake_pipeline/rawdata'
procdata_dir = '../PharmacoDI_snakemake_pipeline/procdata'
gene_sig_dir = '../PharmacoDI_snakemake_pipeline/rawdata/gene_signatures'
drug_meta_file = '../PharmacoDI_snakemake_pipeline/metadata/drugs_with_ids.csv'
output_dir = '../PharmacoDI_snakemake_pipeline/latest'
drug_meta_file = '../PharmacoDI_snakemake_pipeline/metadata/drugs_with_ids.csv'
gct_file = '../PharmacoDI_snakemake_pipeline/rawdata/gene_signatures/metaanalysis/gene_compound_tissue.csv'

if __name__ == '__main__':
    # Step 1 -- works with all PSets
    #pset_dict = pdi.pset_df_to_nested_dict(pdi.read_pset(pset, pset_dir))
    #pdi.build_all_pset_tables(pset_dict, pset, procdata_dir, gene_sig_dir)

    # Step 2
    #pdi.combine_all_pset_tables(procdata_dir, output_dir, drug_meta_file)

    # Step 3
    # pdi.build_compound_synonym_df(drug_meta_file, output_dir)

    # Step 9
    pdi.build_gene_compound_tissue_df(gct_file, output_dir)



import os
import glob
import pandas as pd
import numpy as np


def read_gene_signatures(pset_name, file_path):
    """
    Read all gene signatures for a PSet (to be used in gene_compounds table) from the directory file_path.

    @param pset_name: [`string`] The name of the PSet
    @param file_path: [`string`] The directory that holds all gene signature files
    @return: [`DataFrame`] A dataframe containing all gene signatures for the PSet
    """
    # Find correct pset gene signature CSV file
    pset_file = glob.glob(
        f'{os.path.join(file_path, pset_name, pset_name)}_gene_sig.csv')
    if len(pset_file) == 0:
        raise ValueError(
            f'No PSet gene signatures file named {pset_name} could be found in {file_path}')

    # Read .csv file and return df
    return pd.read_csv(pset_file[0])


def build_gene_compound_tissue_dataset_df(gene_sig_dir, pset_name):
    """
    TODO - ask Chris to explain this table again

    @param gene_sig_dir: [`string`] The file path to the directory containing the gene
        signatures for each PSet
    @param pset_name: [`string`] The name of the PSet
    @return: [`DataFrame`] The gene_compounds table for this PSet, containing all stats (?)
    """
    # If gene signature file doesn't exist, return empty DataFrame
    if not os.path.exists(os.path.join(gene_sig_dir, pset_name)):
        print(
            f'WARNING: gene signature annotations file does not exist for' 
            '{pset_name} in {gene_sig_dir}')
        return None

    # Get gene_sig_df from gene_sig_file
    gene_sig_df = read_gene_signatures(pset_name, gene_sig_dir)

    # Extract relevant columns
    # gene_compound_tissue_dataset = gctd
    gctd_df = gene_sig_df[['gene', 'drug', 'tissue', 'dataset',
        'estimate_analytic', 'lower_analytic', 'upper_analytic', 
        'lower_permutation', 'upper_permutation', 'n', 'pvalue_analytic', 
        'pvalue_permutation', 'df', 'fdr_analytic', 'fdr_permutation',
        'significant_permutation', 'mDataType']].copy()

    # Add missing columns
    gctd_df['sens_stat'] = 'AAC'
    gctd_df['permutation_done'] = 0
    gctd_df[~gctd_df['fdr_analytic'].isna(), 'permutation_done'] = 1

    # Rename foreign key columns
    gctd_df.rename(columns={'gene': 'gene_id', 'drug': 'compound_id',
        'tissue': 'tissue_id', 'dataset': 'dataset_id',
        'estimate_analytic': 'estimate'}, inplace=True)

    # Reorder columns
    return gctd_df[['gene_id', 'compound_id', 'dataset_id', 'tissue_id',
        'estimate', 'lower_analytic', 'upper_analytic', 
        'lower_permutation', 'upper_permutation', 'n', 'pvalue_analytic', 
        'pvalue_permutation', 'df', 'fdr_analytic', 'fdr_permutation',
        'significant_permutation', 'permutation_done', 'sens_stat', 
        'mDataType']]



# from rpy2.robjects import r, pandas2ri

# # TODO: Design PSet class for Python
# def convert_pset_to_py(pset):
#     """
#     Read in .rds file of a PSet and parse each object slot into a nested dictionary replicating the PSet
#     structure.

#     :param pset_file: [string] Path to a PSet .rds file
#     :return: [dict] Nested Python dictionaries holding each item of the R PSet at a key within
#         it the respective dictionary.
#     """
#     pandas2ri.activate()

#     molecular_profiles = rlist_to_dict(pset.slots['molecularProfiles'])
#     if molecular_profiles != []:
#         molecular_profiles = {key: r_summarizedexperiment_to_dict(val) for key, val in molecular_profiles.items()}

#     py_pset = {
#         ## FIXME:: Deal with R objects in annotation slot!
#         'annotation': rlist_to_dict(pset.slots['annotation']),
#         'molecularProfiles': molecular_profiles,
#         'sensitivity': rlist_to_dict(pset.slots['sensitivity']),
#         'perturbation': rlist_to_dict(pset.slots['perturbation']),
#         'drug': pset.slots['drug'],
#         'datasetType': pset.slots['datasetType'],
#         'cell': pset.slots['cell'],
#         'curation': rlist_to_dict(pset.slots['curation'])
#     }
#     return py_pset


# ## Helper functions
# def rlist_to_dict(robj):
#     """
#     Accepts an rpy2 R list object and converts it to a Python dictionary.

#     :param robj: [R 'list'] Named R list object.
#     :return: [dict] R list as a dict with all list elements coerced to Python objects.
#     """
#     # Handle empty list case
#     if not list(robj):
#         return []
#     else:
#         dictionary = try_catch('dict(zip(robj.names, list(robj)))', 'list(robj)', locals())
#         return dictionary


# def try_catch(expr, error, environment):
#     """
#     One line function replicating the tryCatch function from R. Returns the result of `expr` evaluated in `environment`
#     if it successfully executes, otherwise returns `error`

#     :param expr: [string] quoted version of the Python code you wish to execute
#     :param error: [any] what to return if evaluation of `expr` fails
#     :param environment: [dict] dictionary of variables needed to evaluate `expr`. Passed to `globals` argument of `eval`
#     :return: `expr` evaluated with `environment` or `error` if the evaluation returns an error
#     """
#     try:
#         return (eval(expr, environment))
#     except:
#         return (error)


# def r_s4_to_dict(object, recursive=False):
#     """
#     Convert an R S4 object to a Python dictionary, with slot names as keys. Note this will NOT convert any additional
#     S4 objects contained in the object and these will need to be converted manually.

#     :param object: [R 'S4'] object to be coerced to a Python dictionary
#     :return: [dict] with keys named for each slot. Converts R base types (vector, matrix, list, data.frame) to the
#         appropriate Python object, but nested S4 classes must be managed manually.
#     """
#     # Convert S4 object to dictionary
#     dct = {name: try_catch('object.slots[name]', [], {'name': name, 'object': object}) for name in list(object.slotnames())}

#     # Convert any R list
#     dct = {key: rlist_to_dict(val) if 'ListVector' in str(type(val)) else val for key, val in dct.items()}

#     # Recursively convert nested S4 objects to dictionaries
#     if recursive:
#         dct = {key: r_s4_to_dict(val, recursive=True) if 'S4' in str(type(val)) else val for key, val in dct.items()}

#     return dct


# def r_summarizedexperiment_to_dict(object):
#     """
#     Convert an R SummarizedExperiment object into a nested Python dictionary

#     :param object: A R SummarizedExperiment object or a dictionary made from one using the r_s4_to_dict function
#     :return: a nested dictionary with keys corresponding to the items in each slot and values coerced to Python objects
#     """
#     # Convert to dictionary if necessary
#     if 'dict' not in str(type(object)):
#         se = r_s4_to_dict(object)
#     else:
#         se = object

#     # Load R functions into Python
#     as_data_frame = r['as.data.frame']

#     # Deal with metadata
#     metadata = se.get('metadata')
#     if not list(metadata):
#         metadata = metadata
#     else:
#         metadata = {key: r_s4_to_dict(val, recursive=True) if 'S4' in str(type(val)) else val for key, val in metadata.items()}

#     # Extract slot data and convert to Python
#     py_se = {
#         'colData': as_data_frame(se['colData']),
#         ## TODO:: Determine if there is any useful information in the other slots of a SimpleList?
#         'assays': r_s4_to_dict(se.get('assays').slots['data']).get('listData'),
#         'NAMES': se.get('NAMES'),
#         'elementMetadata': as_data_frame(se.get('elementMetadata')),
#         'metadata': metadata
#     }
#     return py_se


# def recursive_class(dct):
#     """
#     Recursively get the class of objects in a nested dictionary tree

#     :param dct: [dict] a dictionary or nested dictionary to get the classes for
#     :return: [dict] a dictionary with the same keys are the `dct` and values as the class at each key
#     """
#     return {key: recursive_class(val) if isinstance(val, dict) else type(val) for key, val in dct.items()}




import os
import glob
import pandas as pd
import numpy as np


def build_primary_pset_tables(pset_dict, pset_name):
    """
    Build the tissue, compound, and gene tables for a PSet and return them
    in a dictionary, with table names as the keys.

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @param pset_name: [`string`] The name of the PSet
    @return: [`dict`] A dictionary of primary tables, with table names as keys
    """
    pset_dfs = {}
    pset_dfs['dataset'] = pd.Series(pset_name, name='name')
    pset_dfs['tissue'] = build_tissue_df(pset_dict)
    pset_dfs['compound'] = build_compound_df(pset_dict)
    pset_dfs['compound_annotation'] = build_compound_annotation_df(pset_dict)
    pset_dfs['cell'] = build_cell_df(pset_dict)
    # Don't make gene table if there are no molecular profiles (TODO - check with chris)
    if 'molecularProfiles' in pset_dict:
        pset_dfs['gene'] = build_gene_df(pset_dict)
        pset_dfs['gene_annotation'] = build_gene_annotation_df(pset_dict)
    return pset_dfs


def build_gene_df(pset_dict):
    """
    Build a table containing all genes in a dataset.

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @return: [`DataFrame`] The gene table
    """
    gene_df = pd.Series([], name='name', dtype='str')
    for mDataType in pset_dict['molecularProfiles']:
        gene_df = gene_df.append(pd.Series(pd.unique(
            pset_dict['molecularProfiles'][mDataType]['rowData']['.features']),
            name='name', dtype='str'))

    # Many ENSEMBL gene IDs have the version (ex. ENST00000456328.2 instead
    # of ENST00000456328); remove all version numbers
    # Keeping version numbers for now!
    #gene_df.replace('\.[0-9]$', '', regex=True, inplace=True)
    
    gene_df.drop_duplicates(inplace=True)
    return gene_df


def build_tissue_df(pset_dict):
    """
    Build a table containing all tissues in a dataset.

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @return: [`DataFrame`] The tissue table
    """
    tissue_df = pd.Series(
        pd.unique(pset_dict['cell']['tissueid']), name='name')
    return tissue_df


def build_compound_df(pset_dict):
    """
    Build a table containing all compounds in a dataset.

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @return: [`DataFrame`] The compound table
    """
    compound_df = pd.Series(pd.unique(pset_dict['drug']['drugid']), name='name')
    return compound_df


def build_gene_annotation_df(pset_dict):
    """
    Build a table mapping each gene in a dataset to its gene annotations.

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @return: [`DataFrame`] A table of all gene annotations, mapped to genes
    """
    gene_annotation_df = pd.DataFrame(
        columns=['gene_id', 'symbol', 'gene_seq_start', 'gene_seq_end'], dtype='str')

    for mDataType in pset_dict['molecularProfiles']:
        df = pset_dict['molecularProfiles'][mDataType]['rowData'].copy()
        # Get gene annotation columns
        cols = ['.features']
        if 'Symbol' in df.columns:
            cols.append('Symbol')

        df = df[cols]
        df.rename(columns={'.features': 'gene_id',
                           'Symbol': 'symbol'}, inplace=True)
        gene_annotation_df = gene_annotation_df.append(df)

    # Remove all ENSEMBL gene id version numbers (ex. ENST00000456328.2 instead of ENST00000456328)
    # For now, we have decided to keep version numbers in the table, they are only dropped in joins
    #gene_annotation_df['gene_id'].replace('\.[0-9]$', '',
    #                                      regex=True, inplace=True)

    gene_annotation_df.drop_duplicates(subset=['gene_id'], inplace=True)

    return gene_annotation_df


def build_compound_annotation_df(pset_dict):
    """
    Build a table mapping each compound in a dataset to its compound annotations.

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @return: [`DataFrame`] A table of all compound annotations, mapped to compounds
    """
    # Make compound_annotations df
    compound_annotation_df = pset_dict['drug'][[
        'rownames', 'smiles', 'inchikey', 'cid', 'FDA']].copy()
    compound_annotation_df.rename(
        columns={'rownames': 'compound_id', 'cid': 'pubchem', 'FDA': 'fda_status'}, inplace=True)

    return compound_annotation_df


# TODO - confirm that you're using the correct cell id
def build_cell_df(pset_dict):
    """
    Build a table containing all the cells in a dataset, mapped to their tissues.

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @return: [`DataFrame`] A table of all cell lines, mapped to tissues
    """
    cell_df = pset_dict['cell'][['cellid', 'tissueid']].copy()
    cell_df.rename(columns={'cellid': 'name',
                            'tissueid': 'tissue_id'}, inplace=True)

    return cell_df



from chembl_webresource_client.new_client import new_client
import multiprocessing as mp
import numpy as np
import pandas as pd
import os


def get_chembl_compound_target_mappings(drug_annotation_file, target_file, drug_target_file):
    """
    Get drug target mappings for all drugs in the drug_annotation files (using standard
    inchi key to map between the file and ChEMBL) and all targets in the target file.
    Write to drug_target file and return the resulting DataFrame.

    :drug_annotation_file: the full file path to the drug_annotation table
    :target_file: the full file path to the target csv file
    :drug_target_file: the full file path where the drug target file will be written
    :return: the ChEMBL drug target DataFrame
    """
    # Load drug annotations table
    if not os.path.exists(drug_annotation_file):
        raise FileNotFoundError(f"The file {drug_annotation_file} does not exist!")
    drug_df = pd.read_csv(drug_annotation_file)

    # Get inchikeys of all drugs (not smiles bc capitalization in ChEMBL is inconsistent..)
    inchikeys = list(pd.unique(drug_df['inchikey'].dropna()))

    # Get ChEMBL drugs with matching inchikeys
    print('Getting all drugs from ChEMBL...')
    chembl_drug_df = pd.concat(parallelize(
        inchikeys, get_drugs_by_inchikey, 50))
    chembl_drug_df = pd.merge(
        drug_df[['compound_id', 'inchikey']], chembl_drug_df, on='inchikey', how='inner')
    chembl_drug_df.drop(columns='inchikey', inplace=True)
    molecule_ids = list(chembl_drug_df['molecule_chembl_id'])

    # Get targets from target_file
    if not os.path.exists(target_file):
        print(f"ERROR: the ChEMBL target file {target_file} doesn't exist!\n"
              "Call get_chembl_targets to generate this file.")
    target_df = pd.read_csv(target_file, index_col=0)
    target_df = target_df[['target_chembl_id',
                           'pref_name', 'target_type', 'accession']].copy()
    target_df.drop_duplicates(inplace=True)
    target_ids = list(pd.unique(target_df['target_chembl_id']))

    # Get mappings between drugs (molecule_ids) and targets (target_ids)
    # TODO: I'm not sure if parallelization actually helps for this API call
    print('Getting drug-target mappings from ChEMBL...')
    drug_target_mappings = parallelize(
        molecule_ids, get_drug_target_mappings, 50, target_ids)
    drug_target_df = pd.concat(drug_target_mappings).drop_duplicates()
    drug_target_df = pd.merge(
        drug_target_df, chembl_drug_df, on='molecule_chembl_id')
    drug_target_df = pd.merge(drug_target_df, target_df, on='target_chembl_id')

    # Reorder columns and write to .csv
    drug_target_df = drug_target_df[['compound_id', 'molecule_chembl_id', 
        'target_chembl_id', 'pref_name', 'accession', 'target_type']].copy()
    drug_target_df.to_csv(drug_target_file)
    return drug_target_df


# TODO: add progress bar to parallelize
def parallelize(queries, operation, chunksize, *args):
    """
    Splits queries into chunks of chunksize and then uses a pool to 
    parallelize operation on the query chunks.

    :queries: list of arguments passed to function (in tuples)
    :operation: function being parallelized
    :chuksize: integer representing how many queries each process should handle
    :return: list of results for each query chunk
    """
    chunked_queries = [queries[i:i+chunksize]
                       for i in np.arange(0, len(queries), chunksize)]
    pool = mp.Pool(mp.cpu_count())

    # If operation requires extra args, use pool.starmap instead of pool.map
    if args:
        for i in range(len(chunked_queries)):
            chunked_queries[i] = (chunked_queries[i], *args)
        results = pool.starmap(operation, chunked_queries)
    else:
        results = pool.map(operation, chunked_queries)

    pool.close()
    return results


def get_drugs_by_inchikey(inchikeys):
    """
    Get all drugs in the ChEMBL database with matching inchikeys.

    :inchikeys: A list of inchikeys
    :return: A dataframe of drugs (including ChEMBL ID and inchikey)
    """
    chembl_drug_df = pd.DataFrame(columns=['inchikey', 'molecule_chembl_id'])
    # Initiate connection to ChEMBL molecule table
    molecule = new_client.molecule
    molecules = molecule.filter(molecule_structures__standard_inchi_key__in=inchikeys).only(
        ['molecule_chembl_id', 'molecule_structures'])
    for mol in molecules:
        inchikey = mol['molecule_structures']['standard_inchi_key']
        chembl_drug_df = chembl_drug_df.append({'inchikey': inchikey,
                                                'molecule_chembl_id': mol['molecule_chembl_id']},
                                               ignore_index=True)
    return chembl_drug_df


def get_drug_target_mappings(molecule_ids, target_ids):
    """
    Retrieves mapping between drugs specified by ChEMBL molecule_ids and targets
    specified by ChEMBL target_ids.

    :molecule_ids: A list of ChEMBL drug IDs
    :target_ids: A list of ChEMBL target IDs
    :return: A DataFrame of drug target mappings (molecule ChEMBL ID and target ChEMBL ID)
    """
    # Initiate connection to ChEMBL activity table
    activity = new_client.activity
    results = activity.filter(molecule_chembl_id__in=molecule_ids, target_chembl_id__in=target_ids,
                              pchembl_value__isnull=False).only(['molecule_chembl_id', 'target_chembl_id'])
    mappings = list(results)
    df = pd.DataFrame(mappings)
    return df



import os
#import PharmacoDI as di
import pandas as pd
import numpy as np
import urllib
import requests
from io import StringIO
from lxml import etree

def get_target_annotations(pset, annot_dir):
    """
    Annotate a the 'TARGET' in the 'drug' slot of a PSet object using mapping from the UniProt idenitifer
    mapping tool API.

    :param pset:
    :param annot_dir:
    :return:
    """
    # Read in drug target annotations and gene annotations
    drug_targets = pd.read_csv(os.path.join(annot_dir, 'drugbank_drug_targets_all.csv'))
    rnaseq_df = pset.get("molecularProfiles").get("Kallisto_0.46.1.rnaseq").get("elementMetadata")

    # Map genes to drugbank drug ids
    genes_to_drugs = pd.merge(drug_targets.loc[:, ['Name', 'Gene Name', 'Drug IDs']],
                              rnaseq_df.loc[:, ['gene_name', 'gene_id']],
                              left_on='Gene Name', right_on='gene_name')

    # Annotate the genes


    # Expand list columns into rows and annotate drugs
    genes_to_drugs['Drug IDs'] = [str.split(ids, '; ') for ids in genes_to_drugs['Drug IDs'].values]
    genes_to_drugs = genes_to_drugs.explode('Drug IDs')


    # Write to disk if necessary.
    file_path = os.path.join(annot_dir, 'drugbank_drug_to_gene_mappings.csv')
    if not os.isfile(file_path):
        pd.write_csv(genes_to_drugs, file_path)
    pass



def query_uniprot_mapping_api(ids, convert_from="ACC+ID", to="ENSEMBL_ID"):
    """
    Query the UniProt mapping API to convert between different gene/protein identifiers.
    Defaults to converting UniProt AC or ID into ENESBLE gene id. The API, however,
    supports a wide range of identifier conversions such as 'GENENAME', "EMBL", and
    "P_ENTREZGENEID".

    Unmatched ids fail silently and will be excluded from the resulting DataFrame. They
    can be retrieved by redoing your query in manually at https://www.uniprot.org/uploadlists/.

    Documentation for other potential conversions are available at:
        https://www.uniprot.org/help/api_idmapping

    :param ids: [`list`, `tuple` or `ndarray`] An iterable sequence type containing the gene/protein
        identifiers as strings.
    :param convert_from: [`string`] The UniProt abbreviation for a format of gene/protein identifier.
        Defaults to 'ACC+ID'.
    :param to: [`string`] The Uniprot abbreviation for the desired gene/protein identifier. Defaults
        to 'ENSEMBL'.
    :return: [`DataFrame`] With the columns 'From' and 'To', mapping from the current id to the selected
        id type based on annotation in the UniProt database.
    """
    # API URL
    url = 'https://www.uniprot.org/uploadlists/'

    # Build the query
    query = ' '.join([str(id) for id in ids])
    params = {
        'from': convert_from,
        'to': to,
        'format': 'tab',
        'query': query
    }
    query_string = urllib.parse.urlencode(params)

    # Encode query and retrieve get request
    query_string = query_string.encode('utf-8')
    req = urllib.request.Request(url, query_string)
    with urllib.request.urlopen(req) as f:
        response = f.read()

    # Decode query and convert to DataFrame
    table_data = StringIO(response.decode("utf-8"))
    uniprot_mapping_df = pd.read_table(table_data, sep="\t")

    return(uniprot_mapping_df)




import os
import warnings
import numpy as np
import re
from datatable import dt, f, g, join, sort, update, fread


# -- Enable logging
from loguru import logger
import sys

logger_config = {
    "handlers": [
        {"sink": sys.stdout, "colorize": True, "format": 
            "<green>{time}</green> <level>{message}</level>"},
        {"sink": f"logs/build_meta_tables.log", 
            "serialize": True, # Write logs as JSONs
            "enqueue": True}, # Makes logging queue based and thread safe
    ]
}
logger.configure(**logger_config)

@logger.catch
def build_gene_compound_tissue_df(gene_compound_tissue_file, gene_file, 
    compound_file, tissue_file, output_dir):
    """
    Build gene_compound_tissue table (description?)

    @param gene_compound_tissue_file: [`str`] Path to the gene signature .csv file.
    @param gene_file: [`str`] Path to the gene table .csv file.
    @param compound_file: [`str`] Path to the compound table .csv file.
    @param tissue_file: [`str`] Path to the tissue table .csv file.
    @param output_dir: [`str`] Path to write the output file to.

    @return [`datatable.Frame`] Writes the 'gene_compound_tissue.csv' file to 
    output_dir then returns the table.
    """
    # -- Check the input files exist
    for file in [gene_compound_tissue_file, gene_file, compound_file, tissue_file]:
        if not os.path.exists(file):
            raise FileNotFoundError(f'Could not find the {file}')

    # -- Read in mapping tables
    gene_dt = fread(gene_file)
    compound_dt = fread(compound_file)
    tissue_dt = fread(tissue_file)

    # -- Read in gene_compound_tissue table
    gct_dt = fread(gene_compound_tissue_file)

    # -- Fix names and assign missing columns
    gct_dt.names = {'Gene': 'gene_id', 'Tissue': 'tissue_id', 
        'Drug': 'compound_id', 'FWER_genes': 'FWER_gene'}
    # Determine missing columns and assign them, so we don't have to change code 
    #>when new columns are addeds
    gct_table_columns = np.asarray(('id', 'gene_id', 'compound_id', 'tissue_id', 
        'estimate', 'lower', 'upper', 'n', 'tstat', 'fstat', 'pvalue', 'df',
        'fdr', 'FWER_gene', 'FWER_compound', 'FWER_all', 'BF_p_all', 'sens_stat', 
        'mDataType', 'tested_in_human_trials', 'in_clinical_trials'))
    gct_missing_columns = np.setdiff1d(gct_table_columns, np.asarray(gct_dt.names))
    for col in gct_missing_columns:
        gct_dt[col] = None
    gct_dt1 = gct_dt[:, list(gct_table_columns)]
    # Sanity check the columns are there
    if not np.all(gct_table_columns == np.asarray(gct_dt1.names)):
        raise ValueError(f'The gene_compound_tissue table',
            ' has missing columns!')

    # -- Map to existing FK ids
    # gene id
    gct_dt1.names = {'gene_id': 'gene_name'}
    gene_dt.names = {'id': 'gene_id', 'name': 'gene_name'}
    gene_dt.key = 'gene_name'
    # NOTE: the g object references the joined tables namespace
    gct_dt1[:, update(gene_id=g.gene_id), join(gene_dt)]

    # check for failed genes
    failed_genes = gct_dt1[dt.isna(f.gene_id), 'gene_name'].to_numpy().flatten()
    if len(failed_genes) > 0:
        raise ValueError(f'Could not find the {file}')

    if (np.any(gct_dt1[:, dt.isna(f.gene_id)].to_numpy())):
        warnings.warn('Some gene_ids in gene_compound_tissue are still NA! Dropping'
            'the missing rows...')
        gct_dt1 = gct_dt1[~dt.isna(f.gene_id), :]

    # compound id
    gct_dt1.names = {'compound_id': 'compound_name'}
    compound_dt.names = {'id': 'compound_id', 'name': 'compound_name'}
    del compound_dt[:, 'compound_uid']
    compound_dt.key = 'compound_name'
    gct_dt1[:, update(compound_id=g.compound_id), join(compound_dt)]

    # tissue id
    gct_dt1.names = {'tissue_id': 'tissue_name'}
    tissue_dt.names = {'id': 'tissue_id', 'name': 'tissue_name'}
    tissue_dt.key = 'tissue_name'
    gct_dt1[:, update(tissue_id=g.tissue_id), join(tissue_dt)]

    ## TODO: Handle failed tissue mappings?

    # -- Sort then assign the primary key column
    ## TODO:: Is there a way to sort by reference?
    gct_dt2 = gct_dt1[:, list(gct_table_columns), 
        sort('gene_id', 'compound_id', 'tissue_id', 'mDataType')]
    gct_dt2[:, update(id=range(1, gct_dt2.nrows + 1))]

    # Sanity check we didn't lose any rows
    if not gct_dt.nrows == gct_dt2.nrows:
        warnings.warn('The compound_gene_tissue table has lost some rows!')

    gct_dt2.to_csv(os.path.join(output_dir, 'gene_compound_tissue.csv'))



## FIXME:: This function is almost identical to the previous one, refactor
##>into a helper for gene_compound instead of copy pasting code
@logger.catch
def build_gene_compound_dataset_df(gene_compound_dataset_file, gene_file, 
    compound_file, dataset_file, output_dir):
    """
    Build gene_compound_dataset table (description?)

    @param gene_compound_dataset_file: [`str`] Path to the gene signature .csv file.
    @param gene_file: [`str`] Path to the gene table .csv file.
    @param compound_file: [`str`] Path to the compound table .csv file.
    @param dataset_file: [`str`] Path to the tissue table .csv file.
    @param output_dir: [`str`] Path to write the output file to.

    @return [`datatable.Frame`] Writes the 'gene_compound_dataset.csv' file to 
        output_dir the returns the table.
    """
    # -- Check the input files exist
    for file in [gene_compound_dataset_file, gene_file, compound_file, dataset_file]:
        if not os.path.exists(file):
            raise FileNotFoundError(f'Could not find the {file}')

    # -- Read in mapping tables
    gene_dt = fread(gene_file)
    compound_dt = fread(compound_file)
    dataset_dt = fread(dataset_file)

    # -- Read in gene_compound_tissue table
    gcd_dt = fread(gene_compound_dataset_file)

    # -- Fix names and assign missing columns
    gcd_dt.names = {'gene': 'gene_id', 'compound': 'compound_id', 
        'dataset': 'dataset_id', 'lower': 'lower_analytic', 
        'upper': 'upper_analytic', 'pvalue': 'pvalue_analytic', 
        'fdr': 'fdr_analytic'}
    del gcd_dt[:, ['significant', 'tissue']]

    # Determine missing columns and assign them, so we don't have to change code 
    #>when new columns are addeds
    gcd_table_columns = np.asarray(('id', 'gene_id', 'compound_id', 'dataset_id', 
        'estimate', 'lower_analytic', 'upper_analytic', 'lower_permutation',
        'upper_permutation', 'n', 'pvalue_analytic', 'pvalue_permutation', 
        'df', 'fdr_analytic', 'fdr_permutation', 'significant_permutation',
        'permutation_done', 'sens_stat', 'mDataType'))
    gcd_missing_columns = np.setdiff1d(gcd_table_columns, np.asarray(gcd_dt.names))
    for col in gcd_missing_columns:
        gcd_dt[col] = None
    gcd_dt1 = gcd_dt[:, list(gcd_table_columns)]
    # Sanity check the columns are there
    if not np.all(gcd_table_columns == np.asarray(gcd_dt1.names)):
        raise ValueError(f'The build_gene_compound_dataset table',
            ' has missing columns!')

    gcd_dt1[:, update(sens_stat='AAC', permutation_done=0)]

    # -- Map to existing FK ids
    # gene id
    gcd_dt1.names = {'gene_id': 'gene_name'}
    # fix gene ids that still have versions
    gsub = np.vectorize(re.sub)
    gcd_dt1[f.gene_name.re_match('ENS.*[.][0-9]*'), 
        update(gene_name=gsub('[.][0-9]*$', '', gcd_dt1[
            f.gene_name.re_match('ENS.*[.][0-9]*'), :]['gene_name'].to_numpy()))]
    gene_dt.names = {'id': 'gene_id', 'name': 'gene_name'}
    gene_dt.key = 'gene_name'
    # NOTE: the g object references the joined tables namespace
    gcd_dt1[:, update(gene_id=g.gene_id), join(gene_dt)]

    # make sure all genes mapped
    failed_genes = np.unique(gcd_dt1[dt.isna(f.gene_id), 'gene_name'] \
        .to_numpy().flatten())
    if len(failed_genes) > 0:
        warnings.warn(f'The genes: {failed_genes} did not map!')


    if (np.any(gcd_dt1[:, dt.isna(f.gene_id)].to_numpy())):
        warnings.warn('Some gene_ids in gene_compound_dataset are still NA!'
            'Dropping the missing rows...')
        gcd_dt1 = gcd_dt1[~dt.isna(f.gene_id), :]

    # compound id
    gcd_dt1.names = {'compound_id': 'compound_name'}
    compound_dt.names = {'id': 'compound_id', 'name': 'compound_name'}
    del compound_dt[:, 'compound_uid']
    compound_dt.key = 'compound_name'
    gcd_dt1[:, update(compound_id=g.compound_id), join(compound_dt)]

    # dataset id
    gcd_dt1.names = {'dataset_id': 'dataset_name'}
    dataset_dt.names = {'id': 'dataset_id', 'name': 'dataset_name'}
    dataset_dt.key = 'dataset_name'
    gcd_dt1[:, update(dataset_id=g.dataset_id), join(dataset_dt)]

    # -- Sort then assign the primary key column
    gcd_dt2 = gcd_dt1[:, list(gcd_table_columns), 
        sort('gene_id', 'compound_id', 'dataset_id', 'mDataType')]
    gcd_dt2[:, update(id=range(1, gcd_dt2.nrows + 1))]

    # Sanity check we didn't lose any rows
    if not gcd_dt.nrows == gcd_dt2.nrows:
        warnings.warn('The gene_compound_dataset table has lost some rows!')

    gcd_dt2.to_csv(os.path.join(output_dir, 'gene_compound_dataset.csv'))

@logger.catch
def build_gene_compound_df(gene_compound_file, gene_file, compound_file, 
    output_dir):
    """
    Build gene_compound table, with pan-cancer and pan-dataset meta-analysis
    results.

    @param gene_compound_file: [`str`] Path to the gene signature .csv file.
    @param gene_file: [`str`] Path to the gene table .csv file.
    @param compound_file: [`str`] Path to the compound table .csv file.
    @param output_dir: [`str`] Path to write the output file to.

    @return [`datatable.Frame`] Writes the 'gene_compound_dataset.csv' file to 
        output_dir the returns the table.
    """
    pass
    # FIXME:: For now this function makes a dummy table since we don't have
    #>real data
    # -- get the all combinations of gene_id x compound_id
    # gene_dt = fread(gene_file)
    # compound_dt = fread(compound_file)
    # gene_id, compound_id = np.meshgrid(gene_dt['id'].to_numpy(), 
    #     compound_dt['id'].to_numpy(), indexing='ij')
    # gene_id = gene_id.flatten()
    # compound_id = compound_id.flatten()

    # # -- build a datatable
    # gc_dt = dt.Frame({
    #     'gene_id': gene_id,
    #     'compound_id': compound_id,
    # })

    # # Determine missing columns and assign them, so we don't have to change code 
    # #>when new columns are addeds
    # gc_table_columns = np.asarray(('id', 'gene_id', 'compound_id', 'estimate', 
    #     'lower', 'upper', 'n', 'tstat', 'fstat', 'pvalue', 'df', 'fdr', 
    #     'FWER_gene', 'FWER_compound', 'FWER_all', 'BF_p_all', 'sens_stat', 
    #     'mDataType', 'tested_in_human_trials', 'in_clinical_trials'))
    # gc_missing_columns = np.setdiff1d(gc_table_columns, np.asarray(gc_dt.names))
    # for col in gc_missing_columns:
    #     gc_dt[col] = None
    # gc_dt1 = gc_dt[:, list(gc_table_columns)]
    # # Sanity check the columns are there
    # if not np.all(gc_table_columns == np.asarray(gc_dt1.names)):
    #     raise ValueError(f'The build_gene_compound_dataset table',
    #         ' has missing columns!')

    # gc_dt1['estimate'] = np.random()
    # gc_dt1['n'] =
    # gc_dt1['pvalue'] =


# -- scripts for testing the function
if __name__ == '__main__':
    if not os.getcwd() == 'PharmacoDI_snakemake_pipeline':
        os.chdir('PharmacoDI_snakemake_pipeline')
    sig_data_dir = os.path.join('rawdata/gene_signatures/metaanalysis')
    data_dir = os.path.join('latest')
    gene_compound_tissue_file = os.path.join(sig_data_dir, 'gene_compound_tissue.csv')
    gene_compound_dataset_file = os.path.join(sig_data_dir, 'gene_compound_dataset.csv')
    gene_file = os.path.join(data_dir, 'gene.csv')
    compound_file = os.path.join(data_dir, 'compound.csv')
    tissue_file = os.path.join(data_dir, 'tissue.csv')
    dataset_file = os.path.join(data_dir, 'dataset.csv')



from .download_psets import *
from .download_canonical_psets import *
from .download_gene_signatures import *
from .get_gene_targets import get_gene_targets
from .get_target_annotations import get_target_annotations, query_uniprot_mapping_api
from .read_pset import pset_df_to_nested_dict, read_pset_file, read_pset
from .build_all_pset_tables import build_all_pset_tables
from .get_chembl_targets import get_chembl_targets
from .get_chembl_compound_targets import get_chembl_compound_target_mappings
from .build_target_tables import build_target_tables
from .combine_pset_tables import combine_all_pset_tables
from .build_synonym_tables import build_cell_synonym_df, build_compound_synonym_df, build_tissue_synonym_df
from .build_cellosaurus import build_cellosaurus_df
from .build_clinical_trial_tables import build_clinical_trial_tables
from .map_genes_to_genomic_coordinates import map_genes_to_genomic_coordinates
from .build_meta_tables import build_gene_compound_tissue_df, build_gene_compound_dataset_df


import pandas as pd
from PharmacoDI.build_all_pset_tables import build_cell_df, build_compound_df, build_tissue_df


def build_dataset_join_dfs(pset_dict, pset_name, primary_dfs={}):
    """
    Builds join tables summarizing the cell lines, tissues, and compounds 
    in this PSet.

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @param pset_name: [`string`] The name of the PSet
    @param primary_dfs: [`pd.DataFrame`] A dictionary containing primary tables (cell,
                                        tissue, compound/compound) from this PSet
    @return: [`dict(string: pd.DataFrame)`] A dictionary of the join tables, with 
                                            table names as keys
    """
    cell_df = primary_dfs['cell'] if 'cell' in primary_dfs else None
    tissue_df = primary_dfs['tissue'] if 'tissue' in primary_dfs else None
    compound_df = primary_dfs['compound'] if 'compound' in primary_dfs else None

    join_dfs = {}
    join_dfs['dataset_cell'] = build_dataset_cell_df(
        pset_dict, pset_name, cell_df)
    join_dfs['dataset_tissue'] = build_dataset_tissue_df(
        pset_dict, pset_name, tissue_df)
    join_dfs['dataset_compound'] = build_dataset_compound_df(
        pset_dict, pset_name, compound_df)
    return join_dfs


def build_dataset_cell_df(pset_dict, pset_name, cell_df=None):
    """
    Builds a join table summarizing the cell lines in this PSet.

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @param pset_name: [`string`] The name of the PSet
    @param cell_df: [`pd.DataFrame`] The cell table for this PSet
    @return: [`pd.DataFrame`] The join table with all cell lines in this PSet
    """
    if cell_df is None:
        cell_df = build_cell_df(pset_dict)

    dataset_cell_df = pd.DataFrame(
        {'dataset_id': pset_name, 'cell_id': cell_df['name']})

    return dataset_cell_df


def build_dataset_tissue_df(pset_dict, pset_name, tissue_df=None):
    """
    Builds a join table summarizing the tissues in this PSet.

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @param pset_name: [`string`] The name of the PSet
    @param tissue_df: [`pd.DataFrame`] The tissue table for this PSet
    @return: [`pd.DataFrame`] The join table with all tissues in this PSet
    """
    if tissue_df is None:
        tissue_df = build_tissue_df(pset_dict)

    dataset_tissue_df = pd.DataFrame(
        {'dataset_id': pset_name, 'tissue_id': tissue_df})

    return dataset_tissue_df


def build_dataset_compound_df(pset_dict, pset_name, compound_df=None):
    """
    Builds a join table summarizing the compounds/compounds in this PSet.

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @param pset_name: [`string`] The name of the PSet
    @param compound_df: [`pd.DataFrame`] The compound/compound table for this PSet
    @return: [`pd.DataFrame`] The join table with all compounds/compounds in this PSet
    """
    if compound_df is None:
        compound_df = build_compound_df(pset_dict)

    dataset_compound_df = pd.DataFrame(
        {'dataset_id': pset_name, 'compound_id': compound_df})

    return dataset_compound_df



import glob
import os
import re
import numpy as np
import pandas as pd
import swifter  # Library to parallelize apply statements automagically

pset_name = 'GDSC_v1'
file_path = os.path.join('data', 'rawdata')
slot_names = ['curation', 'drug', 'molecularProfiles',
              'sensitivity', 'annotation', 'cell']


def read_pset(pset_name, file_path, slot_names=['curation', 'drug', 'molecularProfiles', 'sensitivity', 'annotation', 'cell']):
    """
    Read in all the data associated with a PharmacoSet from the .csv files exported by the writeToCsv method from rPharmacoDI.
    
    @param pset_name: [`string`] Name of the PharmacoSet object as it appears in the directory name for files exported using
        rPharmacoDIs writeToCsv function.
    @param file_path: [`string`] Path to the directory where PharmacoSet data is stored.
    @param slot_names: [`list`] A list of PharmacoSet slots to read in. Defaults to all slots in a PharmacoSet.
    
    @return: [`DataFrame`] A DataFrame with the columns slot, containing the name of the slot data is from, one or more subitem columns, indicating
        the subitem in a slot a data came from, file path, the path data was read from and data, which contains the object associated with the
        selected slot and subitem.
    """
    # Use regex to find the appropriate files
    pset_dir = glob.glob(f'{os.path.join(file_path, pset_name)}_PSet')[0]
    if pset_dir is None:
        raise ValueError(
            f'No PSet directory named {pset_name} could be found in {file_path}')
    
    # List al files for the select PSet, then split on $ to make a DataFrame
    pset_files = pd.Series(os.listdir(pset_dir))
    pset_files_df = pset_files.str.split('$', expand=True)
    
    # Build the file paths to read in data for each row of the DataFrame
    pset_files_df['file_paths'] = [os.path.join(
        pset_dir, file_name) for file_name in pset_files]
    
    # Rename columns
    pset_files_df.columns = [
        'slot', *[f'subitems{i}' for i in range(1, pset_files_df.shape[1] - 1)], 'file_paths']
    
    # Read in PSet data
    pset_files_df['data'] = pset_files_df['file_paths'].swifter.apply(
        read_pset_file)
    
    # Drop file_paths column by reference
    pset_files_df.drop('file_paths', axis='columns', inplace=True)

    # Process id columns to use the proper slot names
    pset_files_df.iloc[:, 0:-1] = pset_files_df.iloc[:, 0:-
                                                     1].apply(lambda col: col.str.replace('.*@|.csv.gz$|.txt', ''))
    
    return pset_files_df


# ---- Helper methods for read_csv
def read_pset_file(file_path):
    """Deal with text files which can't be read in with pd.read_csv"""
    if '.csv.gz' in file_path:
        return pd.read_csv(file_path)
    elif '.txt' in file_path:
        with open(file_path) as f:
            text = [line for line in f]
        return text
    else:
        raise ValueError(
            f'Unsupported file type passed to this function from: {file_path}')


def pset_df_to_nested_dict(df):
    """
    Recrusively turn unique values in the first column of a DataFrame into dictionary keys until only 1 column remains.

    @param df: [`DataFrame`] With one or more ID columns before a data column.
    @return: [`dict`] A nested dict with levels equal to the number of columns of the original DataFrame minus one
    """
    for key_check in pd.unique(df[df.columns[0]]):
        if key_check is not None:
            # Recursively nest the first column as key and remaning columns as values
            return {key: pset_df_to_nested_dict(df.loc[df[df.columns[0]] == key, df.columns[1:]]) if
                    df[1:].shape[1] > 2 else df.loc[df[df.columns[0]] == key, 'data'].values[0] for
                    key in pd.unique(df[df.columns[0]])}
        else:
            # Return the data column if there no no key
            return df.iloc[:, -1].values[0]



# import pandas as pd
# import numpy as np
# import glob
# import os
# import re
#
# def pset_to_db_tables(pset, save_dir=os.path.join("..", "data", "procdata"),
#                       annot_dir=os.path.join("..", "data", "metadata"),
#                       api_url="https://www.orcestra.ca/api/psets/canonical"):
#     """
#     Take in a Python dictionary of a PSet and convert it to database tables for PharmacoDB, with a .csv file for
#     each table saved to `save_dir`.
#
#     :param pset: [dict] a nested dictionary containing data from a PSet, as returned by the `convert_pset_to_py`
#         function.
#     :param save_dir: [string] path to the directory where the database table .csv files should be written
#     :param api_url: [string] URL to fetch available PSets from Orecstra, defaults to current URL
#     :return: [None] writes .csv files to `save_dir`
#     """
#     canonical_names = pd.read_json(api_url).name
#     name = re.sub('_', '.*', pset.get('annotation').get('name')[0])
#
#     # ---- Primary tables ----
#
#     ## ---- dataset
#     dataset = pd.DataFrame.from_dict({
#         "id": [np.where(canonical_names.str.match(name))[0][0] + 1], # Wrap single values in list to make 1 row DF
#         "name": [pset.get('annotation').get('name')[0]],
#     })
#
#
#     ## ---- tissue
#     tissue = pd.DataFrame({
#         'id': np.arange(len(pset.get('cell')['tissueid'].unique())) + 1,
#         'name': pset.get('cell')['tissueid'].unique(),
#     })
#
#
#     ## ---- cell
#     # preprocess cell_info for unique cellid:tissueid combinations and map tissue_id to tissue name
#     cell_info = pset.get('cell')
#     tissue_id_map = dict(zip(tissue.name, tissue.id))
#     cell_info['tissue_id'] = [tissue_id_map[tissue_id] for tissue_id in cell_info.tissueid.to_numpy()] # .to_numpy() should speed up iteration
#
#     # build cell db table
#     cell = cell_info[['cellid', 'tissue_id']].copy()
#     cell.insert(0, 'id', np.arange(1, len(cell_info.cellid) + 1))
#     cell.columns = ['id', 'name', 'tissue_id']
#
#
#     ## ---- compound
#     compound = pd.DataFrame({
#         "id": np.arange(1, len(pset.get('drug').drugid) + 1),
#         "name": pset.get('drug').drugid
#     })
#
#
#     ## ---- gene
#     rnaseq_df = pset.get("molecularProfiles").get("Kallisto_0.46.1.rnaseq").get("elementMetadata")
#     rna_df = pset.get("molecularProfiles").get("rna").get("elementMetadata")
#
#     gene_annots = {re.sub(r"^.*/Gencode\.[^\.]*\.|.csv$", "",  file): pd.read_csv(file) for
#                    file in glob.glob(os.path.join(annot_dir, "Gencode.v33.annotation.*.csv"))}
#
#     # Extract individual annotations from dict
#     feature_genes, feature_transcript, tx2gene = gene_annots.values()
#
#     # Coerce pset datatypes to match gene annotations (some weird stuff happens when converting from R object)
#     rnaseq_df = rnaseq_df.astype(feature_genes.dtypes.to_dict())
#
#     gene = pd.DataFrame({
#         "id": np.arange(1, len(rnaseq_df.rownames) + 1),
#         "name": rnaseq_df.rownames
#     })
#
#     ensg = gene.name.apply(lambda name: re.sub(r"\..*$", "", name))
#
#
#     ## ---- target
#
#
#     target = pd.DataFrame({
#         "id": np.arange(1, len(pset.get("drug")['TARGET'])),
#         "name": pset.get("drug")["TARGET"],
#     })
#
#
#     # ---- Annotation tables ----
#
#     ## ---- cellosaurus
#     cellosaurus = pd.DataFrame({
#         'id': np.arange(1, len(cell.id) + 1), ## FIXME: Can't we just use cell_id as PK?
#         'cell_id': cell.id,
#         'identifier': cell_info['COSMIC.identifier'],
#         'accession': cell_info['Cellosaurus.Accession.id'],
#         'as': np.repeat(None, len(cell.id)),
#         'sy': np.repeat(None, len(cell.id)),
#         'dr': np.repeat(None, len(cell.id)),
#         'rx': np.repeat(None, len(cell.id)),
#         'ww': np.repeat(None, len(cell.id)),
#         'cc': np.repeat(None, len(cell.id)),
#         'st': np.repeat(None, len(cell.id)),
#         'di': np.repeat(None, len(cell.id)),
#         'ox': np.repeat(None, len(cell.id)),
#         'hi': np.repeat(None, len(cell.id)),
#         'oi': np.repeat(None, len(cell.id)),
#         'sx': np.repeat(None, len(cell.id)),
#         'ca': np.repeat(None, len(cell.id)),
#     })
#
#     ## TODO: Parse cellosaurus text file into DataFrame
#
#     compound_annotation = pd.DataFrame({
#         "compound_id": compound.id,
#         "name": compound.name,
#         'smiles': pset.get('drug').smiles,
#         'inchikey': pset.get('drug').inchikey,
#         'pubchem': pset.get('drug').cid,
#         'fda': [0 if fda < 0 else 1 for fda in pset.get('drug').FDA]
#     })
#
#     compound_target = pd.DataFrame({
#         "id": 1,
#         "name": []
#     })
#
#     gene_annotations = pd.DataFrame({
#         "id": gene.id,
#         "symbol": rnaseq_df.gene_name,
#         "gene_seq_start": rnaseq_df.start,
#         "gene_seq_end": rnaseq_df.end,
#     })
#
#     dataset_statistics = pd.DataFrame({
#         "id": 1,
#         "name": []
#     })
#
#
#     # ---- Derived tables ----
#
#     compound_drug = pd.read_csv(glob.glob(os.path.join(annot_dir, "gene_signature", "GDSCV1"))
#
#
#     # ---- Join tables ----
#
#     # ---- Write to .csv ----



import pandas as pd
import wget


def download_psets(names, save_dir, api_url="https://www.orcestra.ca/api/psets/available"):
    """
    Download the specified PSets from the specified api_url

    :param names: [list] Names of PSets to download. Must match the 'names' in the api call JSON.
    :param save_dir: [string] Path to save the PSets in
    :param api_url: [string] URL where available PSets can be retrieved. Defaults to current Orcestra API.
    :return: [None] Downloads PSet .rds files into save_dir using wget.
    """
    pset_df = pd.read_json(api_url)
    names = pd.Series(names)
    if not all(names.isin(pset_df.name)):
        print(names[~names.isin(pset_df.name)] +
                         ' not in names of retrived from api_url')
        raise ValueError(names[~names.isin(pset_df.name)] + 'are not valid pset names')
    pset_df = pset_df[pset_df.name.isin(names)]
    url_dict = pset_df.set_index('name').to_dict()['downloadLink']
    for name, url in url_dict.items():
        print("Downloading", name, "from", url, sep=" ")
        wget.download(url, save_dir + "/" + name + '.rds')

    return None



import pandas as pd
from bs4 import BeautifulSoup
import time
from selenium import webdriver
import requests

ensembl_gene_id = 'ENSG00000140465'


def get_gene_targets(ensembl_gene_id):
    """
    Query the APIs of Ensembl and NCBI to fetch information about drugs targeting a given gene.

    :param gene_symbol: [string] The gene symbol for the desired target gene
    :param ensembl_gene_id: [string] The Ensembl gene id for the desired target gene
    :return: [DataFrame] containing annotations for the specified gene
    """
    # Configure requests to automatically raise errors
    http = requests.Session()

    assert_status_hook = lambda response, *args, **kwargs: response.raise_for_status()
    http.hooks["response"] = [assert_status_hook]

    # Query the APIs
    queries = {'genecards_query':
                   f"https://www.genecards.org/cgi-bin/carddisp.pl?id={ensembl_gene_id.upper()}&idtype=ensembl",
               'ncbi_query': f"https://www.ncbi.nlm.nih.gov/gene/?term={ensembl_gene_id.upper()}"},
    f"http://useast.ensembl.org/Homo_sapiens/Gene/Summary?g={ensembl_gene_id.upper()}"

    api_requests = {key: requests.get(query) for key, query in queries.items()}

    # Retry if 403
    is_403 = [req.status_code == "403" for req in api_requests.values()]
    if (any(is_403)):
        to_update = queries.keys()[is_403]
        updates = [requests.get(query) for query in queries.values()[is_403]]
        api_requests.replace(dict(zip(to_update, updates)))  # Map replacement by value to dict key

    parsed_annotation_data = \
        {key: BeautifulSoup(request.text, 'html.parser') if request.status_code == "200" else None
         for key, request in api_requests.items()}


def scrape_genecards(ensembl_gene_id):
    """
    Use headless Firefox browser to make get requests to Genecards despite their anti-scraping software

    WARNING: Requires Firefox and geckodriver be installed to work!

    :param ensembl_gene_id: [string] The ENSEMBL id for the gene you want to query
    :return: [DataFrame] containing the gene card data
    """
    # Configure to run Chrome/Chromium headless
    options = webdriver.FirefoxOptions()
    #options.add_argument("-disable-extensions")
    #options.add_argument("-disable-gpu")
    #options.add_argument("-no-sandbox")
    #options.add_argument("-headless")

    # Initialize Chrome/Chromium
    driver = webdriver.Firefox(options=options)

    # Build the HTTP query and use the browser to get it
    ensembl_query = f"https://www.genecards.org/cgi-bin/carddisp.pl?id={ensembl_gene_id.upper()}&idtype=ensembl"
    driver.get(ensembl_query)
    expand_table_elements = driver.find_elements_by_xpath("//a[@data-role='show-all']")
    for element in expand_table_elements:
        time.sleep(1)
        element.click()
    page_html = driver.page_source
    del driver

    # Parse the page HTML to a DataFrame
    parsed_html = BeautifulSoup(page_html, 'html.parser')
    genecards_dfs = pd.read_html(str(parsed_html.find_all('table')))




import glob
import os
import re
import numpy as np
import pandas as pd
import datatable as dt

def combine_all_pset_tables(data_dir, output_dir, compound_meta_file):
    """
    Combine all PSet tables into the final PharmacoDB tables.

    @param data_dir: [`string`] The file path to read the PSet tables
    @param output_dir: [`string`] The file path to write the final tables
    @return: [`dict(string: datatable.Frame)`] A dictionary of all some of the
        final tables, with names as keys, to be used for later joins
    """
    print("Combining all PSet tables...")
    join_dfs = combine_primary_tables(data_dir, output_dir, compound_meta_file)
    combine_secondary_tables(data_dir, output_dir, join_dfs)
    combine_experiment_tables(data_dir, output_dir, join_dfs)
    combine_gene_compound_tissue_dataset_tables(data_dir, output_dir, join_dfs)


def combine_primary_tables(data_dir, output_dir, compound_uids_file):
    """
    Build all the primary tables, i.e., tables that require no joins,
    and return them in a dictionary.

    @param data_dir: [`string`] The file path to read the PSet tables
    @param output_dir: [`string`] The file path to write the final tables
    @return: [`dict(string: datatable.Frame)`] A dictionary of all the primary
                                                tables, with names as keys
    """
    # Load, concatenate, and write primary tables to disk
    tissue_df = load_join_write('tissue', data_dir, output_dir)
    gene_df = load_join_write('gene', data_dir, output_dir)
    dataset_df = load_join_write('dataset', data_dir, output_dir)

    # Special handling for compound table: join final table with compounds_with_ids.csv
    compound_df = load_table('compound', data_dir)
    compound_meta_df = dt.fread(compound_uids_file)
    compound_meta_df.names = {'unique.drugid': 'name', 'PharmacoDB.uid': 'compound_uid'}
    compound_meta_df = compound_meta_df[:, ['name', 'compound_uid']]
    compound_meta_df.key = 'name'
    compound_df = compound_df[:, :, dt.join(compound_meta_df)]
    compound_df = write_table(compound_df, 'compound', output_dir)

    # Transform tables to be used for joins
    dfs = {}
    dfs['tissue'] = rename_and_key(tissue_df, 'tissue_id')
    dfs['compound'] = rename_and_key(compound_df, 'compound_id')
    dfs['gene'] = rename_and_key(gene_df, 'gene_id')
    dfs['dataset'] = rename_and_key(dataset_df, 'dataset_id')
    return dfs


def combine_secondary_tables(data_dir, output_dir, join_dfs):
    """
    Build all secondary tables, i.e., all tables that have foreign keys corresponding
    to primary keys of primary tables. The function reads PSet tables from 
    data_dir, concatenates and joins them with tables from join_dfs, and 
    writes them to output_dir.

    @param join_dfs: [`dict(string: datatable.Frame)`] A dictionary of all the primary
                                                    tables, with names as keys
    @param data_dir: [`string`] The file path to read the PSet tables
    @param output_dir: [`string`] The file path to write the final tables
    @return: [`dict(string: datatable.Frame)`] The updated dictionary of join tables
    """
    # Build cell table and add to join_dfs dictionary
    cell_df = load_join_write(
        'cell', data_dir, output_dir, ['tissue'], join_dfs)
    join_dfs['cell'] = rename_and_key(cell_df, 'cell_id')

    # Build compound annotation table
    load_join_write('compound_annotation', data_dir,
                    output_dir, ['compound'], join_dfs, add_index=False)
    # Build gene annotation table
    gene_annot_df = load_table('gene_annotation', data_dir)
    # Remove any rows with no actual annotations (no symbol)
    gene_annot_df = gene_annot_df[dt.f.symbol > "", :]
    # Join the other way so that genes that got cut out are included back in
    gene_annot_df.key = 'gene_id'
    gene_annot_df = join_tables(join_dfs['gene'], gene_annot_df, 'gene_id')
    write_table(gene_annot_df, 'gene_annotation', output_dir, add_index=False)

    # Build join tables
    load_join_write('dataset_cell', data_dir, output_dir,
                    ['dataset', 'cell'], join_dfs, add_index=False)
    load_join_write('dataset_tissue', data_dir, output_dir,
                    ['dataset', 'tissue'], join_dfs, add_index=False)
    # TODO: temporary workaround for dataset_compound until we standardize compound -> compound
    dataset_compound_df = load_table('dataset_compound', data_dir)
    dataset_compound_df = join_tables(
        dataset_compound_df, join_dfs['dataset'], 'dataset_id')
    compound_df = join_dfs['compound'].copy()
    compound_df.names = {'compound_id': 'compound_id'}
    dataset_compound_df = join_tables(
        dataset_compound_df, compound_df, 'compound_id')
    dataset_compound_df = write_table(
        dataset_compound_df, 'dataset_compound', output_dir, add_index=False)

    # Build other secondary tables
    load_join_write('mol_cell', data_dir, output_dir,
                    ['cell', 'dataset'], join_dfs)
    # mol_cells has Kallisto. not sure why. from CTRPv2 (TODO)
    load_join_write('dataset_statistics', data_dir,
                    output_dir, ['dataset'], join_dfs)

    return join_dfs


def combine_experiment_tables(data_dir, output_dir, join_dfs):
    """
    Load and process experiment table, then use it to build the dose response
    and profile tables. Drop the 'name' column from the experiment table before
    writing to a CSV.

    @param join_dfs: [`dict(string: datatable.Frame)`]
    @param data_dir: [`string`] The file path to the PSet tables
    @param output_dir: [`string`] The file path to the final tables
    @return: [`None`]
    """
    # Load all experiments from PSets
    experiment_df = load_join_write('experiment', data_dir, output_dir, [
                                    'cell', 'compound', 'dataset', 'tissue'], join_dfs)
    # Don't write the 'name' column
    experiment_df[:, ['id', 'cell_id', 'compound_id', 'dataset_id', 'tissue_id']].to_csv(
        os.path.join(output_dir, 'experiment.csv'))

    # Rename columns and key experiment table based on experiment name and dataset id
    experiment_df.names = {'name': 'experiment_id'}
    experiment_df = experiment_df[:, ['id', 'experiment_id', 'dataset_id']]
    experiment_df.key = ('dataset_id', 'experiment_id')
    join_dfs['experiment'] = experiment_df

    # Nearly the same code as in load_join_write but has special case handling
    for df_name in ['dose_response', 'profile']:
        df = load_table(df_name, data_dir)
        if df_name == 'profile':
            df[dt.f.IC50 > 1e54, dt.update(IC50=1e54)]
        for fk in ['dataset', 'experiment']:
            df = join_tables(df, join_dfs[fk], fk+'_id')
        del df[:, 'dataset_id']
        write_table(df, df_name, output_dir,
                    add_index=(df_name == 'dose_response'))


def combine_gene_compound_tissue_dataset_tables(data_dir, output_dir, join_dfs):
    gd_df = load_table('gene_compound_tissue_dataset', data_dir)
    tissue_df = join_dfs['tissue']

    # Regularize punctuation in tissue IDs before joining
    # TODO: fix with chris' vectorized version
    for df in [gd_df, tissue_df]:
        tissue_col = df[:, 'tissue_id'].copy().to_pandas()
        tissue_col = tissue_col['tissue_id'].squeeze()
        tissue_col = tissue_col.apply(lambda x: re.sub(' |/', '.', x))  
        if df == tissue_df:
            df.key = None
            df['tissue_id'] = dt.Frame(tissue_col)
            df.key = 'tissue_id'
        else:
            df['tissue_id'] = dt.Frame(tissue_col)

    # Join on gene, compound, dataset, and tissue IDs
    for fk in ['gene', 'compound', 'dataset', 'tissue']:
        print(f'Joining gene_compound_tissue_dataset table with {fk} table...')
        gd_df = join_tables(gd_df, join_dfs[fk], fk+'_id', ignore_versions=True)

    write_table(gd_df, 'gene_compound_tissue_dataset', output_dir)


def load_join_write(name, data_dir, output_dir, foreign_keys=[], join_dfs=None, add_index=True):
    """
    Given the name of a table, load all PSet tables of that name from data_dir,
    join them to any foreign key tables (specified by foreign_keys), and write
    the final combined and joined table to output_dir as a CSV.

    @param name: [`string`] The name of the table
    @param data_dir: [`string`] File path to the directory with all PSet tables
    @param output_dir: [`string`] The file path to the final tables
    @param foreign_keys: [`list(string)`] An optional list of tables that this table
                                            needs to be joined with
    @param join_dfs: [`dict(string: datatable.Frame)`] An optional dictionary of join
        tables (for building out foreign keys); keys are table names
    @param add_index: [`bool`] Indicates whether or not to add a primary key (1-nrows)
        when writing the final table to a .csv
    @return: [`datatable.Frame`] The final combined and joined table
    """
    df = load_table(name, data_dir)

    for fk in foreign_keys:
        print(f'Joining {name} table with {fk} table...')
        if fk not in join_dfs:
            raise KeyError(f'The {name} table has the foreign key {fk}_id but \
                            there is no {fk} table in the join tables dictionary.')
        df = join_tables(df, join_dfs[fk], fk+'_id')

    df = write_table(df, name, output_dir, add_index)
    return df


def load_table(name, data_dir):
    """
    Load all PSet tables with name into a datatable, dropping any duplicate rows.
    
    @param name: [`string`] The name of the table
    @param data_dir: [`string`] File path to the directory with all PSet tables
    @return: [`datatable.Frame`] A datatable containing all rows from all PSets
    """
    print(f'Loading PSet-specific {name} tables from {data_dir}...')
    # Get all files
    files = glob.glob(os.path.join(data_dir, '**', f'*{name}.csv'))
    # Filter so that file path are '{data_dir}/{pset}/{pset}_{name}.csv'
    files = [file_name for file_name in files if re.search(
        data_dir + r'/(\w+)/\1_' + name + '.csv$', file_name)]
    # Read and concatenate tables
    df = dt.rbind(*dt.iread(files, sep=','))
    # Replace any empty strings with None/NA
    df.replace("", None)
    # Drop duplicates (groups by all columns and
    # selects only the first row from each group)
    df = df[0, :, dt.by(df.names)]
    
    return df


def rename_and_key(df, join_col, og_col='name'):
    """
    Prepare df to be joined with other tables by renaming the column
    on which it will be joined and by keying it.

    @param df: [`datatable.Frame`] The table to be keyed.
    @param join_col: [`string`] The name of the join column in other tables
                            (ex. 'tissue_id', 'cell_id', etc.)
    @param og_col: [`string`] The name of the join column in the join table
    @return: [`datatable.Frame`] The keyed and renamed table
    """
    # Rename primary key to match foreign key name (necessary for joins)
    df.names = {og_col: join_col}
    # Only select necessary rows
    df = df[:, ['id', join_col]]
    # Set the key
    df.key = join_col
    return df  # Not necessary? df passed by reference


def join_tables(df1, df2, join_col, ignore_versions=False, delete_unjoined=True):
    """
    Join df2 and df1 based on join_col (left outer join by default).

    @param df1: [`datatable.Frame`] The datatable with the foreign key
    @param df2: [`datatable.Frame`] The join table (ex. tissue datatable)
    @param join_col: [`string`] The name of the columns on which the tables
                            will be joined (ex. 'tissue_id')
    @param ignore_versions: [`bool`]
    @param delete_unjoined: [`bool`] An optional parameter (default=True)
        that lets you keep rows in df1 which didn't join to any rows in df2
    @return [`datatable.Frame`] The new, joined table
    """
    if (join_col not in df1.names) or (join_col not in df2.names):
        print(f'{join_col} is missing from one or both of the datatables passed!',
              'Make sure you have prepared df2 using rename_and_key().')
        return None

    df = df1[:, :, dt.join(df2)]

    # If joining on gene_id, try dropping version numbers and joining again
    # TODO: vectorize this
    if ignore_versions and (join_col == 'gene_id') and (df[dt.isna(df[:, 'id']), :].nrows > 0):
        # Remove version numbers from unmatched gene_ids in gene_compound df
        gene_col = df[dt.isna(df[:, 'id']), join_col]
        gene_col = gene_col.to_pandas()
        gene_col.replace('\.[0-9]+$', '', regex=True, inplace=True)
        df[dt.isna(df[:, 'id']), join_col] = dt.Frame(gene_col)

        # Remove version numbers from gene_ids in gene df
        df2 = df2.to_pandas()
        df2.replace('\.[0-9]+$', '', regex=True, inplace=True)
        df2.drop_duplicates(subset=['gene_id'], inplace=True)
        df2 = dt.Frame(df2)
        df2.key = 'gene_id'

        # Make a copy of all unmatched rows in gene_compound, and delete 'id' col (N/A)
        unmatched_df = df[dt.isna(df[:, 'id']), :].copy()
        del unmatched_df[:, 'id']
        # Join unmatched rows with new gene df, and concat with rows that
        # were matched in the previous join
        df = dt.rbind(df[dt.f.id > 0, :], unmatched_df[:, :, dt.join(df2)])

        # NOTE: datatable doesn't currently have a string replace or regex function
        # This code should be updated when datatable introduced Version 1.0, which
        # will have a re module for these types of regex operations

    # Check to see if any FKs are null
    if df[dt.isna(df[:, 'id']), :].nrows > 0:
        print(f'The following {join_col}s failed to map:')
        unmatched = df[dt.isna(df[:, 'id']), join_col].copy()
        unmatched = unmatched[0, :, dt.by(join_col)]
        print(unmatched)
        if delete_unjoined:
            print(f'Rows with these {join_col}s will be deleted!')
            del df[dt.isna(df[:, 'id']), :]

    # Rename the join col and drop it
    df.names = {join_col: 'drop', 'id': join_col}
    del df[:, 'drop']

    return df


def write_table(df, name, output_dir, add_index=True):
    """
    Add a primary key to df ('id' column) and write it to output_dir
    as a .csv file.

    @param df: [`datatable.Frame`] A PharmacoDB table
    @param name: [`string`] The name of the table
    @param output_dir: [`string`] The directory to write the table to
    @return: [`datatable.Frame`] The indexed PharmacoDB table
    """
    print(f'Writing {name} table to {output_dir}...')

    if add_index:
        # Index datatable
        df = dt.cbind(dt.Frame(id=np.arange(df.nrows) + 1), df)

    df.to_csv(os.path.join(output_dir, f'{name}.csv'))
    return df



import os
import requests
import pandas as pd
import datatable as dt
from urllib3.exceptions import HTTPError
from .get_chembl_compound_targets import parallelize
from .combine_pset_tables import write_table


# TODO: split into more helpers?
def build_clinical_trial_tables(output_dir):
    """
    Build the clinical trial and compound trial tables by querying the
    clinicaltrial.gov API. Queries are made by compound names from the compound
    synonyms table.

    @param output_dir: [`string`] The file path to the directory with all
                                    PharmacoDB tables
    @return: None
    """
    # Load compound synonym table
    compound_file = os.path.join(output_dir, 'compound_synonym.csv')
    compound_df = pd.read_csv(compound_file)[['compound_id', 'compound_name']]

    # Query clinicaltrials.gov API to get clinical trials by compound name
    print('Getting clinical trials from clinicaltrials.gov...')
    all_studies = parallelize(list(compound_df['compound_name']),
                              get_clinical_trials_by_compound_names, 50)
    studies_df = pd.concat(all_studies)

    # Explode list-like columns into separate rows, duplicating the index
    # I only use this because all the fields are returned in arrays for some reason
    object_columns = studies_df.dtypes[studies_df.dtypes ==
                                       'object'].index.values
    for column in object_columns:
        studies_df = studies_df.explode(column)
    # Drop and rename columns
    studies_df.drop(columns='Rank', inplace=True)
    studies_df.rename(columns={'NCTId': 'nct',
                               'SeeAlsoLinkURL': 'link',
                               'OverallStatus': 'status'}, inplace=True)

    # Build clinical trials table
    clin_trial_df = studies_df[['nct', 'link', 'status']].copy()
    clin_trial_df.drop_duplicates('nct', inplace=True)
    clin_trial_df.reset_index(inplace=True, drop=True)
    clin_trial_df['clinical_trial_id'] = clin_trial_df.index + 1

    # Build compound trial table
    compound_trial_df = studies_df[['nct', 'compound_name']].copy()
    compound_trial_df.drop_duplicates(inplace=True)    
    compound_trial_df = pd.merge(compound_trial_df, clin_trial_df, on='nct')
    compound_trial_df = pd.merge(compound_trial_df, compound_df, on='compound_name')

    # Write both tables
    write_table(dt.Frame(clin_trial_df), 'clinical_trial', output_dir, add_index=False)
    write_table(dt.Frame(compound_trial_df[['clinical_trial_id', 'compound_id']]), 'compound_trial', output_dir, add_index=False)


# TODO: shorter names please?
def get_clinical_trials_by_compound_names(compound_names):
    """
    Given a list of compound_names, query the clinicaltrial.gov API iteratively
    to get all trials related to these compounds and return these studies in a table.

    @param compound_names: [`list(string)`] A list of (up to 50) compound names
    @return: [`pd.DataFrame`] A table of all studies, including their rank, study ID,
                                NCT id, recruitment status, link, and compound name.
    """
    all_studies = []
    for compound_name in compound_names:
        min_rank = 1
        max_rank = 1000
        # Make initial API call for this compound
        studies, num_studies_returned, num_studies_found = get_clinical_trials_for_compound(
            compound_name, min_rank, max_rank)

        # If not all studies were returned, make additional calls
        while num_studies_found > num_studies_returned:
            min_rank += 1000
            max_rank += 1000
            more_studies, n_returned, n_found = get_clinical_trials_for_compound(
                compound_name, min_rank, max_rank)
            studies = pd.concat([studies, more_studies])
            num_studies_returned += n_returned
        studies['compound_name'] = compound_name
        all_studies.append(studies)
    return pd.concat(all_studies)


def get_clinical_trials_for_compound(compound_name, min_rank, max_rank):
    """
    Given a compound_name, query the clinicaltrial.gov API to get all trials
    for this compound between min_rank and max_rank (inclusive). Return the 
    studies in a table. If the HTTP request fails or if no studies are
    returned, returns an empty DataFrame.

    @param compound_name: [`string`] A compound name
    @param min_rank: [`int`] The minimum rank of a retrieved study
    @param max_rank: [`int`] The maximum rank of a retrieved study
    @return: [`pd.DataFrame`] A table of retrieved studies, including 
        their rank, study ID, NCT id, recruitment status, and link.
    """
    # Make API call
    base_url = 'https://clinicaltrials.gov/api/query/study_fields'
    params = {
        'expr': compound_name,
        'fields': 'OrgStudyId,NCTId,OverallStatus,SeeAlsoLinkURL',
        'min_rnk': min_rank,
        'max_rnk': max_rank,
        'fmt': 'json'
    }
    r = requests.get(base_url, params=params)

    studies = pd.DataFrame(columns=['Rank', 'NCTId', 'OverallStatus', 'SeeAlsoLinkURL'])

    # Check that request was successful
    if r.status_code != 200:
        print(f'API call for clinical trials related to {compound_name}',
              ' failed for the following reason:\n', r.json()['Error'])
        return studies, 0, 0

    response = r.json()['StudyFieldsResponse']
    if 'StudyFields' in response:
        studies = pd.DataFrame(response['StudyFields'])

    return studies, response['NStudiesFound'], response['NStudiesReturned']



import os
from datatable import Frame

def write_pset_table(pset_df, df_name, pset_name, df_dir):
    """
    Write a PSet table to a CSV file.

    @param pset_df: [`DataFrame`] A PSet DataFrame
    @param pset_name: [`string`] The name of the PSet
    @param df_dir: [`string`] The name of the directory to hold all the PSet tables
    @return [`None`]
    """
    pset_path = os.path.join(df_dir, pset_name)
    # Make sure directory for this PSet exists
    if not os.path.exists(pset_path):
        os.mkdir(pset_path)

    # Convert to datatable Frame for fast write to disk
    pset_df = Frame(pset_df)

    print(f'Writing {df_name} table to {pset_path}...')
    # Use datatable to convert df to csv
    pset_df.to_csv(os.path.join(pset_path, f'{pset_name}_{df_name}.csv'))
    


import os
import requests
import pandas as pd
import numpy as np
from multiprocessing import Pool, cpu_count
import datatable as dt
from PharmacoDI.combine_pset_tables import write_table, rename_and_key, join_tables
from PharmacoDI.get_chembl_compound_targets import parallelize


def build_target_tables(drugbank_file, chembl_file, output_dir, compound_synonym_file):
    """
    Build the target and drug target tables using data from Drugbank
    and ChEMBL.

    @param drugbank_file: [`string`] The full file path to Drugbank targets
    @param chembl_file: [`string`] The full file path to ChEMBL targets
    @param output_dir: [`string`] The directory of all final PharmacoDB tables
    @param compound_synonym_file: [`string`] The file path to the compound synonym table
    @return: None
    """
    # Get Drugbank data
    if not os.path.exists(drugbank_file):
        raise FileNotFoundError(f"The file {drugbank_file} doesn't exist!")
    drugbank_df = pd.read_csv(drugbank_file)
    drugbank_df.rename(columns={'polypeptide.external.identifiers.UniProtKB': 'uniprot_id',
                                'drugName': 'compound_name'}, inplace=True)

    # Get ChEMBL data
    if not os.path.exists(chembl_file):
        raise FileNotFoundError(f"The file {chembl_file} doesn't exist!")
    chembl_df = pd.read_csv(chembl_file, index_col=0)
    chembl_df.rename(columns={'pref_name': 'name',
                              'accession': 'uniprot_id'}, inplace=True)

    target_df = build_target_table(chembl_df, drugbank_df, output_dir)
    build_compound_target_table(chembl_df, drugbank_df, target_df, output_dir, compound_synonym_file)
    build_gene_target_table(chembl_df, drugbank_df, target_df, output_dir)


def build_target_table(chembl_df, drugbank_df, output_dir):
    """
    Using data from the Drugbank and ChEMBL drug target files and
    the UniProt API, build the target table.

    @param chembl_df: [`pd.DataFrame`] The ChEMBL drug target table
    @param drugbank_df: [`pd.DataFrame`] The DrugBank drug target table
    @param output_dir: [`string`] The file path to write the final target table
    @return: [`datatable.Frame`] The target table
    """
    # Combine ChEMBL and Drugbank tables to make target table
    target_df = pd.concat([chembl_df[['name']].copy(),
                           drugbank_df[['name']].copy()])
    target_df.drop_duplicates(inplace=True)

    target_df = write_table(dt.Frame(target_df), 'target', output_dir)
    target_df = rename_and_key(target_df, 'target_id')
    return target_df


def build_compound_target_table(chembl_df, drugbank_df, target_df, output_dir, compound_synonym_file):
    """
    Using data from the Drugbank and ChEMBL drug target files and 
    the target table, build the drug target table.

    @param chembl_df: [`pd.DataFrame`] The ChEMBL drug target table
    @param drugbank_df: [`pd.DataFrame`] The DrugBank drug target table
    @param target_df: [`datatable.Frame`] The target table, keyed
    @param output_dir: [`string`] The file path with all final PharmacoDB tables
    @param compound_synonym_file: [`string`] The file path to the compound synonym table
    @return: [`datatable.Frame`] The drug target table
    """
    # Load compound synonym table from output_dir
    if not os.path.exists(compound_synonym_file):
        raise FileNotFoundError(f"The file {compound_synonym_file} doesn't exist!")
    drug_syn_df = pd.read_csv(compound_synonym_file, dtype={'compound_id': 'int32'})

    # Join drugbank df with drug table (TODO: are we really using drug name to map?)
    drugbank_df = pd.merge(drugbank_df, drug_syn_df, on='compound_name')
    # TODO: from 7521 down to only 122 rows :/

    # Combine ChEMBL and Drugbank tables to make drug target table
    drug_target_df = pd.concat([chembl_df[['name', 'compound_id']].copy(),
                                drugbank_df[['name', 'compound_id']].copy()])
    drug_target_df.rename(columns={'name': 'target_id'}, inplace=True)
    drug_target_df.drop_duplicates(inplace=True)

    # Join with target table
    drug_target_df = dt.Frame(drug_target_df)
    drug_target_df = join_tables(drug_target_df, target_df, 'target_id')
    # Drop rows with no target_id, drop duplicates
    drug_target_df = drug_target_df[dt.f.target_id >= 1, :]
    drug_target_df = drug_target_df[0, :, dt.by(drug_target_df.names)]

    drug_target_df = write_table(
        drug_target_df, 'compound_target', output_dir, add_index=False)
    return drug_target_df


def build_gene_target_table(chembl_df, drugbank_df, target_df, output_dir):
    """
    Build a join table...

    @param chembl_df: [`pd.DataFrame`] The ChEMBL drug target table
    @param drugbank_df: [`pd.DataFrame`] The DrugBank drug target table
    @param target_df: [`datatable.Frame`] The target table, keyed
    @param output_dir: [`string`] The file path with all final PharmacoDB tables
    @return: [`datatable.Frame`] The gene_target table
    """
    # Get target-uniprot mappings from ChEMBL and Drugbank tables
    gene_target_df = pd.concat([chembl_df[['name', 'uniprot_id']].copy(),
                                drugbank_df[['name', 'uniprot_id']].copy()])
    gene_target_df.rename(columns={'name': 'target_id'}, inplace=True)
    gene_target_df.drop_duplicates(inplace=True)

    # Retrieve Uniprot-ENSEMBL gene ID mappings
    uniprot_ids = pd.Series(pd.unique(gene_target_df['uniprot_id']))
    uniprot_ensembl_mappings = pd.concat(
        parallelize(uniprot_ids, map_uniprot_to_ensembl, 1000))
    uniprot_ensembl_mappings.drop_duplicates(inplace=True)

    # Join gene_target table with gene table based on uniprot-ensembl mappings
    gene_target_df = pd.merge(
        gene_target_df, uniprot_ensembl_mappings, on='uniprot_id')
    gene_target_df.drop(columns=['uniprot_id'], inplace=True)

    # Load and key the gene table from output_dir
    gene_file = os.path.join(output_dir, 'gene.csv')
    if not os.path.exists(gene_file):
        raise FileNotFoundError(f"There is no gene file in {output_dir}!")
    gene_df = dt.fread(gene_file, sep=",")
    gene_df = rename_and_key(gene_df, 'gene_id')

    # Join target table with gene table and target table
    gene_target_df = dt.Frame(gene_target_df)
    gene_target_df = join_tables(gene_target_df, gene_df, 'gene_id')
    gene_target_df = join_tables(gene_target_df, target_df, 'target_id')

    # Drop columns that didn't join and drop duplicates
    gene_target_df = gene_target_df[(
        dt.f.target_id >= 1) & (dt.f.gene_id >= 1), :]
    gene_target_df = gene_target_df[0, :, dt.by(gene_target_df.names)]

    gene_target_df = write_table(
        gene_target_df, 'gene_target', output_dir, add_index=False)
    return gene_target_df

# TODO: fix this:
"""
The following gene_ids failed to map:
    | gene_id           
--- + ------------------
  0 | ENSCAFG00000013762
  1 | ENSG00000001630   
  2 | ENSG00000006837   
  3 | ENSG00000008128   
  4 | ENSG00000010219   
  5 | ENSG00000036473   
  6 | ENSG00000062485   
  7 | ENSG00000070087   
  8 | ENSG00000070756   
  9 | ENSG00000073578   
 10 | ENSG00000075886   
 11 | ENSG00000088832   
 12 | ENSG00000099810   
 13 | ENSG00000100197   
 14 | ENSG00000100429   
  … | …                 
253 | ENSG00000288269   
254 | ENSG00000288299   
255 | ENSG00000288359   
256 | ENSG00000288516   
257 | ENSMUSG00000053004

[258 rows x 1 column]

Rows with these gene_ids will be deleted!
"""


def map_uniprot_to_ensembl(uniprot_ids):
    """
    Use the UniProt API to retrieve the ENSEMBL gene IDs
    corresponding to the UniProt IDS.

    @param uniprot_ids: [`list(string)`] A list of UniProt IDs.
    @return: [`pd.DataFrame`] A table mapping UniProt IDs to ENSEMBL gene IDs.
    """
    # Make API call
    params = {
        'from': 'ID',
        'to': 'ENSEMBL_ID',
        'format': 'tab',
        'query': " ".join(uniprot_ids)
    }
    r = requests.get('https://www.uniprot.org/uploadlists/', params=params)

    # Check that request was successful
    r.raise_for_status()

    # Split text into rows (one row per ID) and build df
    # Exclude first row (header)
    gene_id_df = pd.DataFrame(r.text.split("\n")[1:])

    # Split into two columns, UniprotId and EnsemblId
    gene_id_df = gene_id_df[0].str.split(pat="\t", expand=True)

    # Drop empty rows and rename columns
    gene_id_df.dropna(inplace=True)
    gene_id_df.rename(columns={0: 'uniprot_id', 1: 'gene_id'}, inplace=True)

    return gene_id_df



from setuptools import setup, find_packages

setup(name='PharmacoDI',
      version='0.0.9',
      description="Tools for processing R PharmacoSet objects into .csv files of PharmacoDB database tables.",
      url='https://github.com/bhklab/DataIngestion/tree/master/PharmacoDI',
      install_requires=[
        'dask[dataframe]',
        'swifter',
        'datatable',
        'pandas',
        'chembl_webresource_client',
        'wget',
        'bs4',
        'selenium',
        'lxml'
      ],
      author='Evgeniya Gorobets, Christopher Eeles, Benjamin Haibe-Kains',
      author_email='christopher.eeles@uhnresearch.ca, benjamin.haibe.kains@utoronto.ca',
      license='MIT',
      packages=find_packages(),
      zip_safe=False
      )


from datatable import update, f, dt
import numpy as np
import re

# -- Enable logging
from loguru import logger
import sys

logger_config = {
    "handlers": [
        {"sink": sys.stdout, "colorize": True, "format": 
            "<green>{time}</green> <level>{message}</level>"},
        {"sink": f"logs/map_genes_to_genomic_coordinates.log", "serialize": True},
    ],
    "extra": {"user": "someone"}
}
logger.configure(**logger_config)

@logger.catch
def map_genes_to_genomic_coordinates(gene_path, gene_annotation_path, gencode_path):
    """
    Reads in the gene and gene_annotation tables along with a gencode annotation file
    and uses the gencode annotations to assign genomic coordinates to genes in gene_annotations
    before writing the updated table back to disk.

    @param gene_path [`string`] Path to the gene table .csv
    @param gene_annotation_path [`string`] Path to the gene_annotation table .csv
    @param gencode_path [`string`] Path tot he genecode annotation table .csv
    
    @return [None] Modifies gene_annotation table and writes .csv to disk
    """

    # -- Load in the required data
    gene = dt.fread(gene_path)
    gene_annot = dt.fread(gene_annotation_path)
    gencode = dt.fread(gencode_path)

    vsub = np.vectorize(re.sub)
    gencode[:, update(gene_id=vsub('[.][0-9]*$', '', gencode['gene_id'].to_numpy()))]

    # -- Add gene name back to gene_annotation
    gene.key = 'id'
    # join columns need the same name, stupid...
    gene_annot[:, update(id = f.gene_id)]
    gene_annot.key = 'id'
    gene_a = gene_annot[:, :, dt.join(gene)]

    # -- Prepocess the genomic coordinates
    gencode.names = {'gene_id': 'name'}
    gencode =  gencode[:, ['name', 'start', 'end', 'strand', 'seqnames']]
    gencode.key = 'name'
    gene_a.key = 'name'

    # -- Map coordinates to gene_annotations, check that nothing went wrong
    gene_annotation = gene_a[:, :, dt.join(gencode)].copy()
    # sanity check the mappings didn't get messed up
    if not np.all(gene_annotation['name'].to_numpy() == gene['name'].to_numpy()):
        raise ValueError('The gene_annotation table got mangled while trying to map'
            'genomic coordinates!')

    # -- Clean up the table and write to disk
    gene_annotation[:, update(gene_seq_start=f.start, gene_seq_end=f.end, chr=f.seqnames)]
    del gene_annotation[:, ['name', 'id', 'start', 'end', 'seqnames']]

    gene_annotation.to_csv(gene_annotation_path)
    return(gene_annotation)



import pandas as pd
import os
from multiprocessing import Pool, cpu_count
from collections import defaultdict
import datatable as dt
from PharmacoDI.combine_pset_tables import join_tables, write_table, rename_and_key


# Using a default dict because it allows me to append duplicate indexes into a list
# Helper for build_cellosaurus_df
def build_defaultdict(tuple_list):
    def_dict = defaultdict(list)
    for tup in tuple_list:
        def_dict[tup[0]].append(tup[1])
    return def_dict


# Only 1666 rows; try joining with cell synonym df instead ? (TODO)
def build_cellosaurus_df(cellosaurus_path, output_dir):
    """
    Build cellosaurus table.

    @param cellosaurus_path: [`string`] Full file path to cellosaurus file
    @param output_dir: [`string`] The directory to write the cellosaurus table
    @param cell_df: [`datatable.Frame`] The cell table; should be renamed, keyed,
                                        and shouldn't have 'tissue_id' column
    @return: [`datatable.Frame`] The cellosaurus table
    """

    with open(cellosaurus_path) as f:
        file = [line for line in f]

    file = file[55:]
    entries = ''.join(file).split('//\n')
    entry_list = [entry.split('\n') for entry in entries]
    entry_split_list = [[item.split('   ')
                         for item in entry] for entry in entry_list]
    entry_tuple_list = [[(item[0], item[1]) for item in entry if len(
        item) > 1] for entry in entry_split_list]

    pool = Pool(cpu_count() - 1)

    dict_list = pool.map(build_defaultdict, entry_tuple_list)
    dict_list = [dict(item) for item in dict_list]
    dict_list = [{key: '|||'.join(value)
                  for key, value in dct.items()} for dct in dict_list]

    cellosaurus_df = pd.DataFrame(dict_list)
    cellosaurus_df.dropna(axis=1, how='all', inplace=True)

    # Always close your pool or you will have a bunch of processes doing nothing
    pool.close()

    # Drop AG and DT columns (age of donor, date)
    cellosaurus_df.drop(columns=['AG', 'DT'], inplace=True)

    # Rename cols and add cell_id column
    rename_dict = {col: col.lower() for col in cellosaurus_df.columns}
    cellosaurus_df.rename(columns=rename_dict, inplace=True)
    cellosaurus_df.rename(
        columns={'id': 'identifier', 'ac': 'accession'}, inplace=True)
    cellosaurus_df['cell_id'] = cellosaurus_df['identifier']

    # Load cell_df
    cell_path = os.path.join(output_dir, 'cell.csv')
    cell_df = rename_and_key(dt.fread(cell_path, sep=','), 'cell_id')

    # Convert to datatable and join with cell_df
    df = join_tables(dt.Frame(cellosaurus_df), cell_df, 'cell_id')
    df = df[dt.f.cell_id >= 1, :]
    df = df[:, ['cell_id', 'identifier', 'accession', 'as', 'sy',
                'dr', 'rx', 'ww', 'cc', 'st', 'di', 'ox', 'hi', 'oi', 'sx', 'ca']]
    df = write_table(df, 'cellosaurus', output_dir)
    return df


#NOTE: These don't map:
"""
>>> cell_df[~cell_df['id'].isin(ids)]
        id       name  tissue_id
141    142      BT179         19
221    222   COLO_005         19
222    223   COLO_011         19
223    224   COLO_021         19
481    482     HCC812          7
742    743       KRIJ         19
960    961  NCE G-28T          8 This one should map to CVCL_0V15
1183  1184   OESO_009         19
1184  1185   OESO_040         19
1237  1238    PD1503a         19
"""



import pandas as pd
import numpy as np
from PharmacoDI.build_primary_pset_tables import build_cell_df


def build_experiment_tables(pset_dict, pset_name, cell_df=None):
    """
    Build the experiment, dose response, and profile tables for a PSet 
    and return them in a dictionary, with table names as the keys.

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @param pset_name: [`string`] The name of the PSet
    @param cell_df: [`pd.DataFrame`] A table of all the cells in the PSet and their tissues
    @return: [`dict`] A dictionary of experiment-related tables
    """
    pset_dfs = {}
    pset_dfs['experiment'] = build_experiment_df(pset_dict, pset_name, cell_df)
    pset_dfs['dose_response'] = build_dose_response_df(pset_dict, pset_name)
    pset_dfs['profile'] = build_profile_df(pset_dict, pset_name)
    return pset_dfs


def build_experiment_df(pset_dict, pset_name, cell_df=None):
    """
    Build a table with all the experiments of a PSet.

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @param pset_name: [`string`] The name of the PSet 
    @param cell_df: [`pd.DataFrame`] A table of all the cells in the PSet and their tissues
    @return: [`pd.DataFrame`] A table containing all experiments in the dataset
    """
    # Build cell_df if not build already
    ## FIX: Truth value of a DataFrame is ambiguous (Chris)
    if cell_df is None:
        cell_df = build_cell_df(pset_dict)

    # Extract relelvant experiment columns
    experiment_df = pset_dict['sensitivity']['info'][[
        '.rownames', 'cellid', 'drugid']].copy()

    # Rename columns
    experiment_df.rename(
        columns={'.rownames': 'experiment_id', 'cellid': 'cell_id', 'drugid': 'compound_id'}, inplace=True)

    # Add datset_id column
    experiment_df['dataset_id'] = pset_name

    # Add tissue_id column by joining with cells_df
    experiment_df = pd.merge(experiment_df, cell_df[['name', 'tissue_id']],
                             left_on='cell_id', right_on='name', how='left')
    experiment_df = experiment_df[[
        'experiment_id', 'cell_id', 'compound_id', 'dataset_id', 'tissue_id']]

    experiment_df.rename(columns={'experiment_id': 'name'}, inplace=True)

    return experiment_df


# TODO:: Do Python functions pass my reference or copy by default?
# - Python neither; it is 'pass-by-object-reference'
# - When you assign a variable in Python, it creates a name in the namespace; that name contains
#   a reference to the value of that variable (You can use the id() function to see the memory address of an object)
# - When you pass a variable to a function in Python, it creates a new variable, which is a copy of the
#   original variable. There are now two DIFFERENT objects, which contain the memory address of the same value
# - Therefore, if you modify that variable BY REFERENCE within the function, then it will modify the actual value;
#   the original variable will then also be modified (because it points to the same memory location)
# - However, if you reassign the variable name within the function scope, it DOES NOT affect the original variable;
#   instead it modifies the copy of the variable with a reference to a different memory location (the new value)
# - Outside of the function, the original variable still holds a reference to the old memory location; therefore
#   the value of that object is not changed
def build_dose_response_df(pset_dict, pset_name):
    """
    Build a table that, for each experiment in a dataset, lists the compound that was
    tested, the doses in which that compound was administered, and the viability responses 
    corresponding to all the doses.

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @param pset_name: [`string`] The name of the PSet
    @return: [`pd.DataFrame`] A table with all the dose-response mappings for 
                                each experiment
    """
    # Get dose and response info from pset
    dose = pset_dict['sensitivity']['raw.Dose']
    response = pset_dict['sensitivity']['raw.Viability']

    # Rename columns so they can be coerced to int later
    dose_pattern = dose.columns[1][:-1]
    rename_dict = {f'{dose_pattern}{n}': str(
        n) for n in np.arange(1, dose.shape[0])}
    dose.rename(columns=rename_dict, inplace=True)
    response.rename(columns=rename_dict, inplace=True)

    # Reshape the DataFrames using melt to go from 'wide' to 'long'
    dose = dose.melt(id_vars='.exp_id', value_name='dose',
                     var_name='dose_id').dropna()
    dose['dose_id'] = dose.dose_id.astype('int')
    response = response.melt(
        id_vars='.exp_id', value_name='response', var_name='dose_id').dropna()
    response['dose_id'] = response.dose_id.astype('int')

    # Set indices for faster joins (~3x)
    dose.set_index(['.exp_id', 'dose_id'], inplace=True)
    response.set_index(['.exp_id', 'dose_id'], inplace=True)

    # Join on index
    dose_response_df = pd.merge(
        dose, response, left_index=True, right_index=True).reset_index()
    dose_response_df.rename(columns={'.exp_id': 'experiment_id'}, inplace=True)
    dose_response_df.drop(columns=['dose_id'], inplace=True)

    # Add dataset_id for joins
    dose_response_df['dataset_id'] = pset_name

    return dose_response_df


def build_profile_df(pset_dict, pset_name):
    """
    TODO: ask Chris

    @param pset_dict: [`dict`] A nested dictionary containing all tables in the PSet
    @param pset_name: [`string`] The name of the PSet
    @return: [`pd.DataFrame`] A table containing all statistics for each profile 
                                in the PSet (?)
    """
    # Get profiles info
    if 'E_inf' in pset_dict['sensitivity']['profiles'].columns:
        profile_df = pset_dict['sensitivity']['profiles'][[
            '.rownames', 'aac_recomputed', 'ic50_recomputed', 'HS', 'E_inf', 'EC50']].copy()
        profile_df.rename(columns={'.rownames': 'experiment_id', 'aac_recomputed': 'AAC',
                                   'ic50_recomputed': 'IC50', 'E_inf': 'Einf'}, inplace=True)
    else:
        profile_df = pset_dict['sensitivity']['profiles'][[
            '.rownames', 'aac_recomputed', 'ic50_recomputed', 'slope_recomputed', 'einf', 'ec50']].copy()
        profile_df.rename(columns={'.rownames': 'experiment_id', 'aac_recomputed': 'AAC', 'slope_recomputed': 'HS',
                                   'ic50_recomputed': 'IC50', 'einf': 'Einf', 'ec50': 'EC50'}, inplace=True)

    # Add DSS columns - TODO get these values when they are eventually computed
    profile_df['DSS1'] = np.nan
    profile_df['DSS2'] = np.nan
    profile_df['DSS3'] = np.nan

    # Add dataset_id for joins
    profile_df['dataset_id'] = pset_name

    return profile_df[['experiment_id', 'HS', 'Einf', 'EC50', 'AAC',
                       'IC50', 'DSS1', 'DSS2', 'DSS3', 'dataset_id']]
